/* 
 * Copyright (c) 2015 SecureRF, All rights reserved.
 *
 * ----------------------------------------------------------------
 * This computer program is CONFIDENTIAL and a TRADE SECRET of
 * SecureRF.  The receipt or possession of this program does not
 * convey any rights to reproduce or disclose its contents, or to
 * manufacture, use, or sell anything that it may describe, in whole
 * or in part, without the specific written consent of SecureRF. Any
 * reproduction of this program without the express written consent of
 * SecureRF is a violation of the copyright laws and may subject you
 * to criminal prosecution.
 * ----------------------------------------------------------------- 
 */
/*
 * bands.c - band creation, processing, and conversion
 */
#include "aeutil.h"
#include "aelib.h"
#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include <math.h>

typedef struct {
	uint8_t *permutation;
	uint8_t *temp;
	uint8_t *identity_perm;
	uint8_t *delta_perm;
	unsigned int permsize;
	unsigned int count;
	int delta;
	uint8_t index;
} bklctx;

/*+
 * bandAllocateBandList - allocate a band list structure with the number of elements requested.
 *                        the elements will be initialized to 0
 *
 * usage:
 *        bandAllocateBandList (l)
 *
 * args:
 *       l - number of bandwords to allocate
 *
 * returns: NULL - could not allocate
 *          ae_pBandList_t - pointer to allocated bandlist
 *
 * notes:
 *       can be called with l == 0 to create a null band list
 *
 * uses:
 *      ae_malloc - allocate memory
 *      ae_calloc - allocate and clear memory
 *-
 */
ae_pBandList_t
bandAllocateBandList (unsigned int l) {
  ae_pBandList_t p;
  p = (ae_pBandList_t) ae_malloc (sizeof(ae_BandList_t));
  if (p) {
    if (l > 0) {
      p->n = l;
      p->bwl = (ae_pBandWord_t) ae_calloc(l,sizeof(ae_BandWord_t));
    } else {
      p->n = 0;
      p->bwl = NULL;
    }
  }
  return p;
}

/*+
 * bandCloneBandList - make an exact copy of a bandlist structure
 *
 * usage:
 *       bandCloneBandList (bl)
 *
 * args:
 *       bl - bandlist to be cloned
 *
 * returns:
 *       - NULL - could not allocate memory
 *       - pointer to cloned copy
 *
 * notes:
 *       if input pointer is null, then null is returned
 *
 * uses:
 *       ae_malloc
 *       bandCloneBandList
 */ 
ae_pBandList_t
bandCloneBandList (ae_pBandList_t bl) {
  ae_pBandList_t p = NULL;
  unsigned int i;
    
  if (bl) {
    p = (ae_pBandList_t) bandAllocateBandList (bl->n);
    if (p) {
      if (bl->n > 0) {
	for (i=0; i < bl->n; i++) {
	  /* sublist entry */
	  if (bl->bwl[i].bl) {
	    p->bwl[i].bl = bandCloneBandList (bl->bwl[i].bl);
	  } else if (bl->bwl[i].d) {  /* delta */
	    p->bwl[i].d = bl->bwl[i].d;
	  } else {                    /* bandword */
	    p->bwl[i].s[0] = bl->bwl[i].s[0];
	    p->bwl[i].s[1] = bl->bwl[i].s[1];
	    p->bwl[i].e    = bl->bwl[i].e;
	  }
	}
      } else {
	p->n = 0;
	p->bwl = NULL;
      }
    }
  }
  return p;
}

/*+
 * bandAppendToBandList - append a bandlist to an existing bandlist 
 *
 * usage:
 *         bandAppendToBandList (bl,s)
 * 
 * args:
 *        bl - bandlist to be appended to
 *        s - bandlist to append
 *
 * returns:
 *         N/A
 *
 * notes:
 *        Bandlists support the notion of lists of individual lists of band words
 *        hence the bl field in the band word structure is meant to allow for this
 *        and all that this function does is to extend the first band list by one
 *        and add the specified bandlist to the end as another list.
 *
 * uses:
 *       ae_realloc
 *       bandCloneBandList
 *-
 */
void
bandAppendToBandList (ae_pBandList_t bl, ae_pBandList_t s) {
  bl->n++;
  bl->bwl = (ae_pBandWord_t) ae_realloc (bl->bwl, bl->n*sizeof(ae_BandWord_t));
  bl->bwl[bl->n-1].bl = bandCloneBandList (s);
  bl->bwl[bl->n-1].d = 0;
  bl->bwl[bl->n-1].s[0] = 0;
  bl->bwl[bl->n-1].s[1] = 0;
  bl->bwl[bl->n-1].e = 0;
    
}

/*+
 * bandAppendBandListToBandList - append one bandlist to another at the same level.
 *
 * usage: bandAppendBandListToBandList (bl, s)
 *
 * args:
 *      bl - target band list
 *      s  - band list to append onto the end of bl
 *
 * returns:
 *     N/A
 *
 * notes:
 *      bandlists are multi level, this routine will join one bandlist to another
 *      at the same level. The assumption is that the list being joined is a single
 *      level list. Otherwise, strange things may happen when one goes to free the
 *      result of this function.
 *
 * uses:
 *      ae_realloc
 *
 *-
 */
void
bandAppendBandListToBandList (ae_pBandList_t bl, ae_pBandList_t s) {
  int i;
  if (s->n == 0) {
    return;
  }
  bl->bwl = ae_realloc (bl->bwl, (bl->n+s->n)*sizeof(ae_BandWord_t));
  for (i=0; i < s->n; i++) {
    bl->bwl[bl->n+i] = s->bwl[i];
  }
  bl->n += s->n;  /* Set new length */
}

/*+
 * bandAppendToBandListEx - append a bandlist to an existing bandlist, flattening the list
 *                          by one level.
 *
 * usage:
 *         bandAppendToBandListEx (bl,s)
 * 
 * args:
 *        bl - bandlist to be appended to
 *        s - bandlist to append
 *
 * returns:
 *         N/A
 *
 * notes:
 *        Bandlists support the notion of lists of individual lists of band words
 *        hence the bl field in the band word structure is meant to allow for this
 *        and all that this function does is to extend the first band list by one
 *        and add the specified bandlist to the end as another list.
 *
 * uses:
 *       ae_realloc
 *       bandCloneBandList
 *-
 */
void
bandAppendToBandListEx (ae_pBandList_t bl, ae_pBandList_t s) {
  unsigned int i,j,len;
  len = bl->n+s->n;
  bl->bwl = (ae_pBandWord_t) ae_realloc (bl->bwl, len*sizeof(ae_BandWord_t));
  for (i=bl->n,j=0; i < len; i++,j++) {
    bl->bwl[i].bl = bandCloneBandList (s->bwl[j].bl);
    bl->bwl[i].d    = s->bwl[j].d;
    bl->bwl[i].s[0] = s->bwl[j].s[0];
    bl->bwl[i].s[1] = s->bwl[j].s[1];
    bl->bwl[i].e    = s->bwl[j].e;
  }
  bl->n = len;
}

/*+
 * bandFreeBandList - free a bandlist allocated with bandAllocateBandList.
 *
 * usage:
 *       bandFreeBandList (bl)
 *
 * args:
 *       bl - pointer to bandlist to free
 *
 * returns: N/A
 *
 * notes:
 *
 * uses:
 *      ae_free - free memory allocated by ae_calloc, ae_malloc, or ae_realloc
 *-
 */
void
bandFreeBandList (ae_pBandList_t bl) {
  unsigned int i;

  if (bl != NULL) {
    for (i=0; i < bl->n; i++) {
      bandFreeBandList (bl->bwl[i].bl);
    }
    if (bl->bwl)
      ae_free (bl->bwl);
    ae_free (bl);
  }
}

/*
 * bandCompareBandList - compare two bandlists for equality
 *
 * usage: bandCompareBandList (orig, new)
 *
 * args:
 *       orig - base band list for comparison
 *       new  - band list to compare
 *
 * returns:
 *       0 - band lists are equal
 *       1 - band lists are different
 *
 * notes:
 *
 * uses:
 *
 *-
 */
unsigned int
bandCompareBandList (ae_pBandList_t orig, ae_pBandList_t new) {
  unsigned int i;
  unsigned int sts;
  if (orig == NULL || new == NULL)
    return 1;
  if (orig->n != new->n)
    return 1;
  for (i=0; i < orig->n; i++) {
    if (orig->bwl[i].bl) {
      sts = bandCompareBandList (orig->bwl[i].bl,new->bwl[i].bl);
      if (sts)
	return 1;
    } else if (orig->bwl[i].d &&
	       orig->bwl[i].d != new->bwl[i].d) {
      return 1;
    } else {
      if (orig->bwl[i].s[0] != new->bwl[i].s[0] ||
	  orig->bwl[i].s[1] != new->bwl[i].s[1])
	return 1;
    }
  }
  return 0;
}

/*+
 * bandAllocateIntList - allocate a list of integers
 * 
 * usage:
 *        bandAllocateIntList (n)
 *
 * args:
 *      n - number of integer slots to allocate
 *
 * returns:
 *         - NULL - could not allocate memory
 *         - pointer to allocated list
 *
 * notes:
 *       use bandFreeIntList to free the allocated list
 *
 * uses:
 *      ae_malloc
 *      ae_calloc
 *-
 */
ae_pIntList_t
bandAllocateIntList (unsigned int l) {
  ae_pIntList_t p;
  p = (ae_pIntList_t) ae_malloc (sizeof(ae_IntList_t));
  if (p) {
    if (l > 0) {
      p->n = l;
      p->items = (unsigned int *) ae_calloc (l, sizeof (unsigned int));
    } else {
      p->n = 0;
      p->items = NULL;
    }
  }
  return p;
}

/*+
 * bandAllocateIdentityPermutation -- allocate a new Identity Permutation
 *
 * usage:
 *	bandAllocateIdentityPermutation(n)
 *
 * args:
 *	n -- braid index
 *
 * returns:
 *	newly allocated and initialized permutation
 *
 * notes:
 *	use bandFreePermutation to free to allocated list
 *
 * uses:
 *	bandAllocatePermutation()
 */
ae_pPermutation_t
bandAllocateIdentityPermutation (unsigned int n)
{
  ae_pPermutation_t p = bandAllocatePermutation(n);
  int i;
  for (i = 0; i < n; i++) {
    p->items[i] = i+1;
  }
  return p;
}

/*+
 * bandFreeIntList - free a allocated integer list
 * 
 * usage:
 *        bandFreeIntList (p)
 *
 * args:
 *        p - pointer to integer list to free
 *
 * returns:
 *        N/A
 *
 * notes:
 *
 * uses:
 *      ae_free
 *
 *-
 */
void
bandFreeIntList (ae_pIntList_t p) {
  if (p != NULL) {
    if (p->n > 0)
      ae_free (p->items);
    ae_free (p);
  }
}

/*
 * bandCompareIntList - compare two Integer Lists for equality
 *
 * usage: bandCompareIntList (orig, new)
 *
 * args:
 *       orig - base integer list for comparison
 *       new  - interger list to compare
 *
 * returns:
 *       0 - integer lists are equal
 *       1 - interger lists are different
 *
 * notes:
 *
 * uses:
 *
 *-
 */
unsigned int
bandCompareIntList (ae_pIntList_t a, ae_pIntList_t b) {
  int i;
  if (!a || !b)
    return 1;
  if (a->n != b->n)
    return 1;
  for (i = 0; i < a->n; i++) {
    if (a->items[i] != b->items[i])
      return 1;
  }
  return 0;
}

/*+
 *
 * bandPrintIntList - print an integer list datatype
 *
 * usage:
 *          bandPrintIntList (fd, p, comment)
 *
 * args:
 *          fd - file descriptor to use
 *          p - pointer to list to print
 *          comment - comment to print prior to the list
 *
 * returns:
 *         N/A
 *
 * notes:
 *        if list is a null pointer then nothing is outpu
 *
 * uses:
 *      fprintf
 *
 *-
 */
void
bandPrintIntList (FILE * fd, ae_pIntList_t p, char * comment) {
  unsigned int i;
  if (p != NULL) {
    fprintf (fd, "List for %s, size: %d: { ",comment,p->n);
    for (i=0; i < p->n; i++) {
      if (i != 0) {
	fprintf (fd,", %d", p->items[i]);
      } else {
	fprintf (fd,"%d",p->items[i]);
      }
    }
    fprintf (fd,"}\n");
  }
}

/*+
 *
 * bandPrintIntListFlat - print an integer list datatype with no extra stuff
 *
 * usage:
 *          bandPrintIntList (fd, p)
 *
 * args:
 *          fd - file descriptor to use
 *          p - pointer to list to print
 *
 * returns:
 *         N/A
 *
 * notes:
 *        if list is a null pointer then nothing is outpu
 *
 * uses:
 *      fprintf
 *
 *-
 */
void
bandPrintIntListFlat (FILE * fd, ae_pIntList_t p) {
  int i;
  if (p != NULL) {
    fprintf (fd, "{");
    for (i=0; i < p->n; i++) {
      if (i) {
	fprintf (fd,",%d", p->items[i]);
      } else {
	fprintf (fd,"%d",p->items[i]);
      }
    }
    fprintf (fd,"}\n");
  }
}
void                   
bandPrintIntListFlatFile (const char *filename, ae_pIntList_t p) {
  FILE *fd;
  int i;
  fd = fopen(filename,"w");
  if (fd) {
    fprintf(fd,"%d\n",p->n);
    for (i=0;i<p->n;i++) 
      fprintf (fd,"%d\n", p->items[i]);
    fclose(fd);
  }
  return;
}

/*+
 *
 * bandAppendToIntList - add one item to the end of an integer list
 *
 * usage:
 *         bandAppendToIntList (p, v)
 *
 * args:
 *         p - pointer to int list
 *         v - value to append
 *
 * returns:
 *          N/A
 *
 * notes:
 *       if intlist pointer is null, nothing occurs
 *
 * uses:
 *       ae_realloc
 *
 *-
 */
void
bandAppendToIntList (ae_pIntList_t p, unsigned int v) {

  if (p != NULL) {
    p->n++;
    p->items = (unsigned int *)ae_realloc (p->items,p->n*sizeof(unsigned int));
    p->items[p->n-1] = v;
  }
}

/*+
 * bandAllocateListOfIntLists - allocate a list of individual integer lists
 *
 * usage:
 *         bandAllocateListOfIntLists (n)
 *
 * args:
 *         n - number of slots to allocate. 0 is ok.
 *
 * returns:
 *        - NULL - could not allocate memory
 *        - pointer to list of int lists allocated
 *
 * notes:
 *
 * uses:
 *       ae_malloc
 *       ae_calloc
 *
 *-
 */
ae_pListOfIntLists_t
bandAllocateListOfIntLists (unsigned int n) {
  ae_pListOfIntLists_t l;

  l = (ae_pListOfIntLists_t) ae_malloc (sizeof(ae_ListOfIntLists_t));
  if (n > 0) {
    l->n = n;
    l->p = (ae_pIntList_t *) ae_calloc (n,sizeof(ae_pIntList_t));
  } else {
    l->n = 0;
    l->p = NULL;
  }
  return l;
}

/*+
 * bandAppendToListOfIntLists - append to the end of the list of lists
 *
 * usage:
 *         bandAppendToListOfIntLists (l,p)
 *
 * args:
 *         l - list to append to
 *         p - int list to append
 *
 * returns:
 *         N/A
 *
 * notes:
 *        NULL pointer to list results in nothing being changed.
 *        item becomes part of appended list, must not be freed except in
 *        the context of the larger list.
 *
 * uses:
 *       ae_realloc
 *-
 */
void
bandAppendToListOfIntLists (ae_pListOfIntLists_t l, ae_pIntList_t p) {
  if (l != NULL) {
    l->p = (ae_pIntList_t *) ae_realloc (l->p, (l->n+1) * sizeof(ae_pIntList_t));
    l->p[l->n++] = (ae_pIntList_t) p;
  }
}
/*+
 * bandCloneIntList - make an exact copy of an IntList
 *
 * usage: bandCloneIntList (s)
 *
 * args:
 *      s - pointer to IntList to clone
 *
 * returns:
 *      pointer to new IntList or NULL if memory could not be allocated
 *
 * notes:
 *
 * uses:
 *      bandAllocateIntList
 *-
 */
ae_pIntList_t
bandCloneIntList (ae_pIntList_t s) {
  int i;
  ae_pIntList_t d;

  d = bandAllocateIntList (s->n);
  if (!d)
    return NULL;

  for (i=0; i < s->n; i++) {
    d->items[i] = s->items[i];
  }
  return d;
}

/*
 * bandSortIntList -- sort the integer list
 *
 * usage:
 *	bandSortIntList(il)
 *
 * args:
 *	il  -- integer list to sort
 *
 * returns:
 *	nothing -- il is sorted in-place
 *
 * notes:
 *	modifies the il passed in
 *
 * uses:
 *
 */
void
bandSortIntList(ae_pIntList_t il)
{
  int i, j;
  unsigned int tmp;

  if (!il)
    return;

  // This is a stupid implementation, but for small lengths should be fine */
  for (i = 1; i < il->n; i++) {
    for (j = i; j > 0; j--) {
      if (il->items[j] < il->items[j-1]) {
	tmp = il->items[j];
	il->items[j] = il->items[j-1];
	il->items[j-1] = tmp;
      }
    }
  }
}

/*+
 * bandFindPermutationFixedPoints - find the fixed points of the permutation
 *
 * usage:
 *         bandFindPermutationFixedPoints (p)
 *
 * args:
 *         p - input permutation
 *
 * returns:
 *         int-list of the fixed permutation points
 *
 * notes:
 *         result is 1-indexed
 *
 * uses:
 *       ae_calloc
 *-
 */
ae_pIntList_t
bandFindPermutationFixedPoints(ae_pPermutation_t perm)
{
  ae_pIntList_t result = NULL;
  int i, j = 0;

  if (!perm)
    return NULL;

  result = bandAllocateIntList(perm->n);
  for (i = 1; i <= perm->n; i++) {
    if (perm->items[i-1] == i)
      result->items[j++] = i;
  }

  result->n = j;
  return result;
}

/*+
 * bandAppendIntListsToIntLists - append an IntLists to another IntLists
 *
 * usage: bandAppendIntListsToIntLists (d, s)
 *
 * args:
 *      d - destination
 *      s - source
 *
 * returns: N/A
 *
 * notes:
 *
 * uses:
 *       ae_realloc
 *       bandCloneIntLists
 *-
 */
void
bandAppendIntListsToIntLists (ae_pListOfIntLists_t d, ae_pListOfIntLists_t s) {
  int i;
  d->p = ae_realloc (d->p, (d->n+s->n)*sizeof(ae_pIntList_t));
  for (i=0; i < s->n; i++) {
    d->p[d->n+i] = bandCloneIntList (s->p[i]);
  }
  d->n += s->n;
}


/*+
 * bandFreeListOfIntLists - free a list of int lists
 *
 * usage:
 *        bandFreeListOfIntLists (l, flags)
 *
 * args:
 *       l - pointer to list to free
 *       flags - if 1, then free each individual int list first
 *
 * returns:
 *         N/A
 *
 * notes:
 *
 * uses:
 *       bandFreeIntList
 *       ae_free
 *
 *-
 */
void
bandFreeListOfIntLists (ae_pListOfIntLists_t l, unsigned int flags) {
  unsigned int i;
  if (l != NULL) {
    if (flags) {
      for (i=0; i < l->n; i++) {
	bandFreeIntList (l->p[i]);
	l->p[i] = NULL;
      }
    }
    if (l->p)
      ae_free(l->p);
    ae_free (l);
  }
}

/*+
 * bandPrintListOfIntLists - print a list of integer lists to the specified file descriptor
 *
 * usage:
 *         bandPrintListOfLists (fd, l, comment)
 *
 * args:
 *       fd - file descriptor to write to
 *       l - pointer to list of integer lists
 *       comment - character string comment
 *
 * returns: N/A
 *
 * notes:
 *
 * uses:
 *       fprintf
 *-
 */
void 
bandPrintListOfIntLists (FILE * fd, ae_pListOfIntLists_t l, char * comment) {
  unsigned int i,j;

  fprintf (fd,"ListOfIntLists %s, length %d\n", comment, l->n);
  if (l->n) {
    for (i=0; i< l->n; i++) {
      fprintf (fd,"    [%d](%d) {",i,l->p[i]->n);
      for (j=0; j<l->p[i]->n;j++) {
	fprintf (fd,"%d ",l->p[i]->items[j]);
      }
      fprintf (fd,"}\n");
    }
  }
}

/*+
 * bandPrintListOfIntListsFlat - print a list of integer lists to the specified file descriptor
 *                             - only print cycles of lengths >= 2
 *
 * usage:
 *         bandPrintListOfListsFlat (fd, l)
 *
 * args:
 *       fd - file descriptor to write to
 *       l - pointer to list of integer lists
 *
 * returns: N/A
 *
 * notes:
 *
 * uses:
 *       fprintf
 *-
 */
void 
bandPrintListOfIntListsFlat (FILE * fd, ae_pListOfIntLists_t l) {
  unsigned int i,j;

  //  fprintf (fd,"ListOfIntLists %s, length %d\n", comment, l->n);
  if (l->n) {
    for (i=0; i< l->n; i++) {
      if(l->p[i]->n > 1) {
	fprintf (fd,"("); //,i,l->p[i]->n);
	for (j=0; j<l->p[i]->n;j++) {
	  if(j)
	    fprintf (fd,",%d",l->p[i]->items[j]);
	  else
	    fprintf (fd,"%d",l->p[i]->items[j]);
	}
	fprintf (fd,")");
      }
    }
    fprintf(fd, "\n");
  }
}

/*+
 * printBandList - print out the values of the band list in standard form, {{s1,s2},e}
 *
 * usage:
 *        printBandList (fd, bl, comment, width)
 *
 * args:
 *      fd - output file to use to print bandlist
 *      bl - the bandlist to print
 *      comment - comment to associate with the bandlist
 *      width - number of bandwords to print per line
 *
 * returns: N/A
 *
 * notes:
 *
 * uses:
 *      fprintf - formatted print
 *-
 */
void
bandPrintBandList (FILE * fd, ae_pBandList_t bl, unsigned char * comment, unsigned int width) {
  unsigned int i;
    
  if (fd == NULL)
    return;

  if (bl == NULL)
    return;

  if (bl->n == 0)
    return;

  if (bl->bwl == NULL)
    return;

  if (comment != NULL) 
    fprintf (fd, "Dump of %s at 0x%p, %d elements\n",comment, bl, bl->n);

  for (i=0; i < bl->n; i++) {
    if (bl->bwl[i].bl) {
      bandPrintBandList (fd,bl->bwl[i].bl,NULL,bl->n);
    } else if (bl->bwl[i].d) {
      fprintf (fd, "{{delta},%d}\n",bl->bwl[i].d);
    } else {
      fprintf (fd, "{{%d,%d},%d} ",bl->bwl[i].s[0],bl->bwl[i].s[1],bl->bwl[i].e);
    }
  }
  fprintf (fd,"}\n");

}

/*+
 * dumpBandList - dump out the values of the band list one character per line
 *
 * usage:
 *        dumpBandList (fd, bl)
 *
 * args:
 *      fd - output file to use to print bandlist
 *      bl - the bandlist to print
 *
 * returns: N/A
 *
 * notes:
 *
 * uses:
 *      fprintf - formatted print
 *-
 */
void
bandDumpBandList(FILE * fd, ae_pBandList_t bl)
{
  int i;

  for (i=0; i < bl->n; i++) {
    if (bl->bwl[i].bl) {
      bandDumpBandList (fd,bl->bwl[i].bl);
    } else if (bl->bwl[i].d) {
      fprintf (fd, "\"delta\"\n%d\n",bl->bwl[i].d);
    } else {
      fprintf (fd, "%d\n%d\n%d\n",bl->bwl[i].s[0],bl->bwl[i].s[1],bl->bwl[i].e);
    }
  }
}

/*+
 * bandPrintBandListFlat - print out the values of the band list in flat,
 *                         standard, nested form:
 *
 *    [len
 *    {d,u}
 *    {{s1,s2},e}
 *    ...
 *    [len
 *    ...
 *    ]
 *    ]
 *
 * usage:
 *        bandPrintBandListFlat (fd, bl)
 *
 * args:
 *      fd - output file to use to print bandlist
 *      bl - the bandlist to print
 *
 * returns: N/A
 *
 * notes:
 *
 * uses:
 *      fprintf - formatted print
 *-
 */
void
bandPrintBandListFlat(FILE *fd, ae_pBandList_t bl)
{
  unsigned int i;

  if (fd == NULL || bl == NULL || bl->n == 0 || bl->bwl == NULL)
    return;

  /* Print the length of this word */
  fprintf(fd, "[%d\n", bl->n);

  /* Print all the items on this list */
  for (i=0; i < bl->n; i++) {
    if (bl->bwl[i].bl) {
      /* Recurse... */
      bandPrintBandListFlat(fd, bl->bwl[i].bl);
    } else if (bl->bwl[i].d) {
      fprintf(fd, "{d,%d}\n", bl->bwl[i].d);
    } else {
      fprintf(fd, "{{%d,%d},%d}\n", bl->bwl[i].s[0], bl->bwl[i].s[1], bl->bwl[i].e);
    }
  }
  fprintf (fd,"]\n");
}

/*+
 * bandPrintBandListReallyFlat - print out the values of the band list in flat,
 *                               standard form without nesting:
 *
 *    [len
 *    {d,u}
 *    {{s1,s2},e}
 *    ]
 *
 * usage:
 *        bandPrintBandListReallyFlat (fd, bl)
 *
 * args:
 *      fd - output file to use to print bandlist
 *      bl - the bandlist to print
 *
 * returns: N/A
 *
 * notes:
 *
 * uses:
 *      fprintf - formatted print
 *-
 */
static void
bandPrintBandListReallyFlatIter(FILE *fd, ae_pBandList_t bl)
{
  unsigned int i;

  /* Print all the items on this list */
  for (i=0; i < bl->n; i++) {
    if (bl->bwl[i].bl) {
      /* Recurse... */
      bandPrintBandListReallyFlatIter(fd, bl->bwl[i].bl);
    } else if (bl->bwl[i].d) {
      fprintf(fd, "{d,%d}\n", bl->bwl[i].d);
    } else {
      fprintf(fd, "{{%d,%d},%d}\n", bl->bwl[i].s[0], bl->bwl[i].s[1], bl->bwl[i].e);
    }
  }
}

void
bandPrintBandListReallyFlat(FILE *fd, ae_pBandList_t bl)
{
  unsigned int len;

  if (fd == NULL || bl == NULL || bl->n == 0 || bl->bwl == NULL)
    return;

  len = bandBandListLength(bl);

  /* Print the length of this word */
  fprintf(fd, "[%d\n", len);
  bandPrintBandListReallyFlatIter(fd, bl);
  fprintf (fd,"]\n");
}
 
/*+
 * bandBandListLength - return the total length of the bandlist
 *
 * usage:  bandBandListLength (bl)
 *
 * args:
 *         bl - band list
 *
 * returns:
 *         total number of band words in the band list
 *
 * notes:
 *
 * uses:
 *-
 */
unsigned int
bandBandListLength(ae_pBandList_t bl)
{
  unsigned int len;
  unsigned int i;

  if (bl == NULL || bl->n == 0 || bl->bwl == NULL)
    return 0;

  len = bl->n;
  for (i=0; i < bl->n; i++) {
    if (bl->bwl[i].bl)
      len += bandBandListLength(bl->bwl[i].bl)-1;
  }

  return len;
}

/*+
 * bandDelta - produce a delta band list of the specified size
 *
 * usage:  bandDelta (n)
 *
 * args:
 *         n - size of list
 *
 * returns:
 *         NULL - could not allocate memory
 *         pointer to created delta bandlist
 *
 * notes:
 *
 * uses:
 *      bandAllocateBandList - allocates an empty bandlist
 *-
 */
ae_pBandList_t
bandDelta(unsigned int n) {
  ae_pBandList_t bl = NULL;
  unsigned int i;

  if (n > 0) {
    bl = bandAllocateBandList (n-1);
    for (i=0; i < n-1; i++) {
      bl->bwl[i].s[0] = n-i;
      bl->bwl[i].s[1] = n-i-1;
      bl->bwl[i].e = 1;
    }
  }
  return bl;
}

/*+
 * bandArtinToBand - convert an Artin list to a Band List
 *
 * usage:
 *        bandArtinToBand (braidlist)
 *
 * args:
 *        braidlist - input braidlist in Artin form
 *
 * returns:
 *        NULL -could not allocate memory
 *        pointer to created band list
 *
 * notes:
 *
 * uses:
 *        bandAllocateBandList - allocate an empty band list
 *
 *-
 */
ae_pBandList_t
bandArtinToBand (ae_plist_t braidlist) {
  ae_pBandList_t bl = NULL;
  unsigned int i;

  if (braidlist && braidlist->len) {
    bl = bandAllocateBandList(braidlist->len);
    for (i=0; i < braidlist->len; i++) {
      bl->bwl[i].s[0] = braidlist->b[i].l+1;
      bl->bwl[i].s[1] = braidlist->b[i].l;
      bl->bwl[i].e    = braidlist->b[i].u;
    }
  }
  return bl;
}

/*+
 * bandGenToArtinWord - convert a band generator to an Artin Word
 *
 * usage:
 *        bandGenToArtinWord (bw)
 *
 * args:
 *        bw - band word generator to convert
 *
 * returns:
 *        NULL - could not allocate memory
 *        pointer to Artin Word
 *
 * notes:
 *
 * uses:
 *       braidAllocateBraidList
 *       braidFreeBraidList
 *
 *-
 */
ae_plist_t
bandGenToArtinWord(ae_pBandWord_t bw) {
  ae_plist_t leftSide, rightSide,aw = NULL;
  unsigned int i, t, s;
  char e;
  unsigned int ai=0,lc=0,rc=0;

  if (bw != NULL) {
    t = bw->s[0];
    s = bw->s[1];
    e = bw->e;
    leftSide = braidAllocateBraidlist(t);
    for (i=s+1; i < t; i++) {
      leftSide->b[lc].l = s+t-i;
      leftSide->b[lc].u = 1;
      lc++;
    }
    rightSide = braidAllocateBraidlist(t);
    for (i=s+1; i<t;i++) {
      rightSide->b[rc].l = i;
      rightSide->b[rc].u = (ae_braiddata_t) -1;
      rc++;
    }
    /*
     * join(leftside,mid,rightside)
     */
    aw = braidAllocateBraidlist(lc+rc+1);
    for (i=0; i<lc; ai++,i++) {
      aw->b[ai].l = leftSide->b[i].l;
      aw->b[ai].u = leftSide->b[i].u;
    }

    aw->b[ai].l   = s;
    aw->b[ai++].u = e;

    for (i=0; i<rc; ai++,i++) {
      aw->b[ai].l = rightSide->b[i].l;
      aw->b[ai].u = rightSide->b[i].u;
    }
    braidFreeBraidlist (&leftSide);
    braidFreeBraidlist (&rightSide);
  }
  return aw;
}
 
/*+
 * bandWordToArtinWord - convert a bandword list to a braidword list
 *
 * usage:
 *        bandWordToArtinWord (bl)
 *
 * args:
 *        bl - pointer to band word list
 *
 * returns:
 *        NULL - could not allocate memory
 *        pointer to created braid list
 *
 * notes:
 *       
 * uses:
 *       braidAllocateBraidList
 *       bandGenToArtinWord
 *       braidFreeBraidlist
 *
 *-
 */
ae_plist_t
bandWordToArtinWord(ae_pBandList_t bl) {
  ae_plist_t aw = NULL;
  ae_plist_t temp;
  unsigned int i,len=0,j,k;
  if (bl && bl->n) {
    /*
     * run through once to figure out how big the list will be.  there
     * must be a better way to do this.  this uses a lot of
     * malloc/free for what should be a simple computation.
     */
    for (i=0; i < bl->n; i++) {
      temp = bandGenToArtinWord(&bl->bwl[i]);
      len += temp->len;
      braidFreeBraidlist(&temp);
    }
    aw = braidAllocateBraidlist(len);
    for (i=0,j=0; i < bl->n; i++) {
      temp = bandGenToArtinWord(&bl->bwl[i]);
      for (k=0; k < temp->len; k++) {
	aw->b[j].l   = temp->b[k].l;
	aw->b[j++].u = temp->b[k].u;
      }
      braidFreeBraidlist(&temp);
    }
  }
  return aw;
}
 
/*+
 * bandInverseOf - determine the inverse of the input band word list
 *
 * usage:
 *        bandInverseOf(bl)
 *
 * args:
 *        bl - band word list
 *
 * returns:
 *        NULL - could not allocate memory
 *        pointer to new band word list now inverted
 *
 * notes:
 *
 * uses:
 *        bandAllocateBandList
 *-
 */
ae_pBandList_t
bandInverseOf (ae_pBandList_t bl) {
  ae_pBandList_t ibl = NULL;
  unsigned int i;

  if (bl != NULL) {
    ibl = bandAllocateBandList (bl->n);
    for (i=0; i < bl->n; i++) {
      ibl->bwl[i].s[0] = bl->bwl[bl->n-i-1].s[0];
      ibl->bwl[i].s[1] = bl->bwl[bl->n-i-1].s[1];
      ibl->bwl[i].e    = -bl->bwl[bl->n-i-1].e;
    }
  }
  return ibl;
}

/*+
 * bandDescendingCycle - create a descending sequence of band generators
 *
 * usage:
 *        bandDescendingCycle (l,r)
 *
 * args:
 *        l - the upper limit of the band generators
 *        r - the starting value of the generators
 *
 * returns:
 *        - NULL - could not allocate memory
 *        - 0 length list, list is too short to create sequence
 *        - pointer to allocated sequence of band generators
 *
 * notes:
 *
 * uses:
 *       bandAllocateBandList
 DescendingCycle[L_, R_] := 
 Module[{Right, Answer}, If[R == 1, Right = 2, Right = R];
 If[L == 1, Answer = { },  
 Answer = Reverse[Table[{{m, m - 1}, 1}, {m, Right, L}]]];
 Answer];  *-
*/
ae_pBandList_t
bandDescendingCycle(int l, int r) {
  ae_pBandList_t bl;
  unsigned int right;
  unsigned int i,j;
  int len;

  if (r == 1) {
    right = 2;
  } else {
    right = r;
  }
  if (l <= 1) {
    bl = bandAllocateBandList (0);
  } else {
    /* generate band in reverse order */
    len = l-right+1;
    if (len <= 0) {
      len = 0;
    }
    bl   = bandAllocateBandList (len);
    for (i=right,j=l-right; i <= l; i++,j--) {
      bl->bwl[j].s[0] = i;
      bl->bwl[j].s[1] = i-1;
      bl->bwl[j].e = 1;
    }
  }
  return bl;
}
 
/*+
 * bandInverseBandGenRewrite - replaces the inverse of a band generator with
 *                              delta raised to the -1 power and a sequence of
 *                              four, possibly empty, cycles.
 *
 * usage:
 *         bandInverseBandGenRewrite (bandgen, n)
 *
 * args:
 *       bandgen - band word generator
 *       n - max value for subscripts
 *
 * returns:
 *         - NULL - could not allocate memory
 *         - pointer to band word list
 *
 * notes:
 *       delta raised to a negative power is a special form of a band word. You'll need to
 *       read the paper to understand more on how its used. 
 *
 * uses:
 *      bandAllocateBandList
 *      bandFreeBandList
 *-
 */
ae_pBandList_t
bandInverseBandGenRewrite(ae_pBandWord_t bandgen, unsigned int n) {
  ae_pBandList_t bl = NULL;
  ae_pBandList_t firstWord,secondWord,thirdWord,fourthWord;
  int t,s;
  unsigned int i=0,j;

  if (bandgen->e == 1) {
    bl = bandAllocateBandList (1);
    bl->bwl[0].s[0] = bandgen->s[0];
    bl->bwl[0].s[1] = bandgen->s[1];
    bl->bwl[0].e    = bandgen->e;
  } else {
    t = bandgen->s[0];
    s = bandgen->s[1];
    firstWord = bandDescendingCycle(n,t+1);

    if (s == 1) {
      secondWord = bandAllocateBandList (0);
    } else {
      secondWord = bandAllocateBandList (1);
      secondWord->bwl[0].s[0] = t;
      secondWord->bwl[0].s[1] = s-1;
      secondWord->bwl[0].e    = 1;
    }

    thirdWord = bandDescendingCycle(s-1,2);

    fourthWord = bandDescendingCycle(t-1,s+1);
    /*
     * create the result band word generator list
     */
    bl = bandAllocateBandList (firstWord->n+secondWord->n+
			       thirdWord->n+fourthWord->n+1);
    /*
     * Join (delta,firstWord,secondWord,thirdWord,fourthWord)
     */
    bl->bwl[i++].d = (short) -1; /* delta */
    /* firstWord */
    for (j=0; j<firstWord->n; j++,i++) {
      bl->bwl[i].s[0] = firstWord->bwl[j].s[0];
      bl->bwl[i].s[1] = firstWord->bwl[j].s[1];
      bl->bwl[i].e    = firstWord->bwl[j].e;
    }
    bandFreeBandList (firstWord);
    /* secondWord */
    for (j=0; j<secondWord->n; j++,i++) {
      bl->bwl[i].s[0] = secondWord->bwl[j].s[0];
      bl->bwl[i].s[1] = secondWord->bwl[j].s[1];
      bl->bwl[i].e    = secondWord->bwl[j].e;
    }
    bandFreeBandList (secondWord);
    /* thirdWord */
    for (j=0; j<thirdWord->n; j++,i++) {
      bl->bwl[i].s[0] = thirdWord->bwl[j].s[0];
      bl->bwl[i].s[1] = thirdWord->bwl[j].s[1];
      bl->bwl[i].e    = thirdWord->bwl[j].e;
    }
    bandFreeBandList (thirdWord);
    /* fourthWord */
    for (j=0; j<fourthWord->n; j++,i++) {
      bl->bwl[i].s[0] = fourthWord->bwl[j].s[0];
      bl->bwl[i].s[1] = fourthWord->bwl[j].s[1];
      bl->bwl[i].e    = fourthWord->bwl[j].e;
    }
    bandFreeBandList (fourthWord);
  }
  return bl;
}
 
/*+
 * bandModifiedMod - modified modulo, if x is 0 or n, then return n,
 *                   otherwise return x modulo n
 *
 * usage:
 *        bandModifiedMod (x, n)
 *
 * arguments:
 *        x - the item to take the modulo of
 *        n - the base for modulo
 *
 * returns:
 *         if x == 0 or x == n, returns n
 *         x modulo n
 *
 * notes:
 *
 * uses:
 *
 *-
 */
unsigned int
bandModifiedMod(unsigned int x, unsigned int n) {

  if (x == 0) {
    return n;
  }
  if (x == n) {
    return n;
  }
  return x % n;
}
 
/*+
 * bandLocalTauInverse - compute the inverse of the specified band generator.
 *
 * usage:
 *        bandLocalTauInverse (bw, n, tinvbw)
 *
 * arguments:
 *        bw - original band word
 *        n - max value for subscripts
 *        tinvbw - tau inverted band word
 *
 * returns:
 *        no return value from this function
 *
 * notes:
 *
 * uses:
 *       bandModifiedMod - modified modulo 
 *-
 */
void
bandLocalTauInverse(ae_pBandWord_t bw, unsigned int n,  ae_pBandWord_t tinvbw) {
  unsigned int tt,ss;
  tinvbw->e = bw->e;
  tt = bandModifiedMod (bw->s[0]-1,n);
  ss = bandModifiedMod (bw->s[1]-1,n);
  tinvbw->s[0] = ss >= tt?ss:tt;
  tinvbw->s[1] = ss <= tt?ss:tt;
}
 
/*+
 * bandTauInverse - compute tau inverse of specified band word list
 *
 * usage:
 *        bandTauInverse (bl, n)
 *
 * arguments:
 *        bl - band word list
 *        n - max value for subscripts
 *
 * returns:
 *        - NULL - could not allocate memory
 *        - pointer to inverted band word list
 *
 * notes:
 *
 * uses:
 *       bandAllocateBandList
 *       bandLocalTauInverse
 *-
 */
ae_pBandList_t
bandTauInverse(ae_pBandList_t bl, unsigned int n) {
  ae_pBandList_t tinvbl = NULL;
  unsigned int i;

  tinvbl = bandAllocateBandList (bl->n);
  for (i=0; i < bl->n; i++) {
    bandLocalTauInverse (&bl->bwl[i],n,&tinvbl->bwl[i]);
  }
  return tinvbl;
}
 
/*+
 * bandIteratedTauInverse - computes tau to the N bandword, tau to the -N bandword where
 *                          N is the number of iterations
 *
 * usage:
 *        bandIteratedTauInverse (iterations, bl, n)
 *
 * arguments:
 *        iterations - number of iterations to run
 *        bl - band list to iterate over
 *        n - maximum value for subscripts
 *
 * returns:
 *       - NULL - could not allocate memory
 *       - pointer to inverted band list
 *
 * notes:
 *
 * uses:
 *      bandTauInverse
 *      bandCloneBandList
 *      bandFreeBandList
 *-
 */
ae_pBandList_t
bandIteratedTauInverse(unsigned int iterations, ae_pBandList_t bl, unsigned int n) {
  ae_pBandList_t t, temp;
  unsigned int i;

  t = bandCloneBandList (bl);
  for (i=0; i < iterations; i++) {
    temp = bandTauInverse (t, n);
    bandFreeBandList (t);
    t = temp;
  }
  return t;
}
 
/*+
 * bandPositiveSubword - computes the i' th subword which appears when rewriting a word in
 *                       band generators after replacing negative powers of generators with 
 *                       Delta(-1) times a positive word
 * 
 * usage:
 *        bandPositiveSubword (i, bl, n)
 *
 * args:
 *        i - subword position
 *        bl - band word list
 *        n - max value of subscripts
 *
 * returns:
 *        - NULL - could not allocate memory
 *        - pointer to allocated list
 *
 * notes:
 *
 * uses:
 *       bandAllocateBandList
 *       bandInverseBandGenRewrite
 *       bandAllocateBandList
 *       bandFreeBandList
 *-
 */
ae_pBandList_t
bandPositiveSubword(unsigned int i, ae_pBandList_t bl, unsigned int n) {
  ae_pBandList_t psw = NULL, temp;
  unsigned int j;

  if (bl->bwl[i].e > 0) {
    psw = bandAllocateBandList (1);
    psw->bwl[0].s[0] = bl->bwl[i].s[0];
    psw->bwl[0].s[1] = bl->bwl[i].s[1];
    psw->bwl[0].e    = bl->bwl[i].e;
  } else {
    temp = bandInverseBandGenRewrite (&bl->bwl[i],n);
    psw = bandAllocateBandList (temp->n-1);
    for (j=0; j < psw->n; j++) {
      psw->bwl[j].s[0] = temp->bwl[j+1].s[0];
      psw->bwl[j].s[1] = temp->bwl[j+1].s[1];
      psw->bwl[j].e    = temp->bwl[j+1].e;
    }
    bandFreeBandList (temp);
  }
  return psw;
}
 
/*+
 * bandFirstBKLRewrite - computes the rewritten BandWord as a negative power of Delta times 
 *                       a product of canonical factors
 *
 * usage:
 *         bandFirstBKLRewrite (bl, n)
 *         bandFirstBKLRewriteWithCarry (bl, n, len, delta)
 *
 * args:
 *        bl - band word list
 *        n  - maximum value for subscripts
 *        len - length of bandword (don't trust bl->n)
 *        delta - carryover from previous loop
 *
 * returns:
 *        - NULL - could not allocate memory
 *        - pointer to resultant band word list
 *
 * notes:
 *
 * uses:
 *       bandAllocateBandList
 *       bandIteratedTauInverse
 *       bandCloneBandList
 *       bandFreeBandList
 *
 *-
 */
static ae_pBandList_t
bandFirstBKLRewriteWithCarry(ae_pBandList_t bl, unsigned int n, unsigned int len, int delta) {
  ae_pBandList_t psw, answer;
  int i, theta;

  if (!bl || !n || !len)
    return NULL;

  /*
   * Start with the carry-over delta (or 0)
   * Iterate from the end of the braid to the beginning.  Do the following:
   * - calculate theta for the ith item
   * - compute the positive subword
   * - compute the iterated tau inverse using the current delta and save
   *   the result to the answer bandlist
   * - delete the positive subword
   * - add the current theta to the current delta
   *
   * once we reach the beginning of the input, prepend the delta element
   * to the first bandword in the answer bandlist
   */
  answer = bandAllocateBandList(len+1);
  delta = -delta;
  for (i=len-1; i >=0; i--) {
    theta = THETA(i, bl);
    psw = bandPositiveSubword(i, bl, n);
    answer->bwl[i+1].bl = bandIteratedTauInverse(delta, psw, n);
    bandFreeBandList(psw);
#ifdef BKL1_DBG
    fprintf(stderr, "iti(psw(%d),delta=%d)\n", i, delta);
#endif
    delta += theta;
  }
  answer->bwl[0].d = -delta;

  return answer;
}

ae_pBandList_t
bandFirstBKLRewrite(ae_pBandList_t bl, unsigned int n) {
  unsigned int len;
  int delta = 0;

  if (!bl)
    return NULL;

  /* Determine the length of the bandlist; handle a trailing delta */
  len = bl->n;
  if (bl->bwl[len-1].d) {
    len--;
    delta = bl->bwl[len].d;
  }

  return bandFirstBKLRewriteWithCarry(bl, n, len, delta);
}
/*+
 * bandFirstBKLToArtin -  converts a negative power of Delta times a product of 
 *                        canonical factors (output of FirstBKLRewrite) to Artin form as a
 *                        product of Artin generators
 *
 * usage:
 *        bandFirstBKLToArtin (bl,n)
 *
 * args:
 *      bl - band list to convert
 *      n - maximum value for subscripts
 *
 * returns:
 *     - NULL - could not allocate memory
 *     - pointer to list of Artin generators
 *
 * notes:
 *
 * uses:
 *      bandDelta
 *      bandInverseOf
 *      bandWordToArtinWord
 *      braidCopyBraidList
 *      braidAllocateBraidlist
 *      braidFreeBraidlist
 *      bandGenToArtinWord
 *      bandFreeBandList
 *-
 */
ae_plist_t
bandFirstBKLToArtin (ae_pBandList_t bl, unsigned int n) {
  int u = -bl->bwl[0].e,i;
  int L = bl->n;
  unsigned int j,k;
  ae_pBandList_t dbl,idbl;
  ae_plist_t answer,t1,t2;
  if (bl->bwl[0].d) {
    u = -bl->bwl[0].d;
  }
  dbl = bandDelta(n);
  idbl = bandInverseOf (dbl);
  answer = bandWordToArtinWord (idbl);
  /*
   * raise delta to the minus power
   */
  if (u > 1) {
    t1 = braidCopyBraidlist (answer);
    for (i = 1; i < u; i++) {
      t2 = braidAllocateBraidlist (answer->len+t1->len);
      for (j=0; j < answer->len; j++) {
	t2->b[j].l = answer->b[j].l;
	t2->b[j].u = answer->b[j].u;
      }
      for (k=0; k < t1->len; j++,k++) {
	t2->b[j].l = t1->b[k].l;
	t2->b[j].u = t1->b[k].u;
      }
      braidFreeBraidlist (&answer);
      answer = t2;
    }
    braidFreeBraidlist (&t1);
  }

  if (L > 1) {
    for (i=1; i < (int)bl->n; i++) {
      t1 = bandWordToArtinWord (bl->bwl[i].bl);
      t2 = braidAllocateBraidlist (answer->len+t1->len);
      for (j=0; j<answer->len; j++) {
	t2->b[j].l = answer->b[j].l;
	t2->b[j].u = answer->b[j].u;
      }
      for (k=0; k<t1->len;k++,j++) {
	t2->b[j].l = t1->b[k].l;
	t2->b[j].u = t1->b[k].u;
      }
      braidFreeBraidlist (&answer);
      answer = t2;
      braidFreeBraidlist (&t1);
    }
  }
  bandFreeBandList (dbl);
  bandFreeBandList (idbl);

  return answer;
}
 
/*+
 * bandMultiplyPermutations - gives the product of the permutations a and b in the symmetric 
 *                            group on n symbols
 * 
 * usage:
 *        bandMultiplyPermutations(a,b)
 *
 * args:
 *        a - first permutation
 *        b - second permutation
 *
 * returns:
 *        - NULL - cannot allocate memory
 *        - pointer to the product list
 *
 * notes:
 *
 * uses:
 *      bandAllocatePermutation
 *-
 */
ae_pPermutation_t
bandMultiplyPermutations(ae_pPermutation_t a, ae_pPermutation_t b) {
  ae_pIntList_t result;
  unsigned int i;

  result = bandAllocatePermutation (a->n);
  if (result) {
    for (i=0; i < a->n; i++) {
      result->items[i] = a->items[(b->items[i]-1)];
    }
  }
  return result;
}

/*+
 * bandPermuteIntList - permutes one permutation using the another
 *
 * usage:
 *        bandPermuteIntList(a,b)
 *
 * args:
 *        a - permutation to be permuted
 *        b - permutation to use to permute a
 *
 * returns:
 *        - NULL - cannot allocate memory
 *        - pointer to the permuted permutation
 *
 * notes:
 *
 * uses:
 *      bandAllocatePermutation
 *-
 */
ae_pIntList_t
bandPermuteIntList(const ae_pIntList_t a, ae_pPermutation_t b) {
  ae_pIntList_t result;
  unsigned int i;

  if (!a || !b)
    return NULL;
  result = bandAllocateIntList (a->n);
  if (result) {
    for (i=0; i < a->n; i++) {
      result->items[b->items[i]-1] = a->items[i];
    }
  }
  return result;
}

/*+
 * bandIntListIndexOf - finds the index of a member in an intlist
 *
 * usage:
 *        bandIntListIndexOf(a,m)
 *
 * args:
 *        a - int list to search
 *        b - member to search for
 *
 * returns:
 *        - -1 - cannot find member
 *        - 0-index into list if member is found
 *
 * notes:
 *
 * uses:
 *
 *-
 */
int
bandIntListIndexOf(const ae_pIntList_t a, unsigned int m)
{
  int i;

  if (!a)
    return -1;

  for (i=0; i < a->n; i++) {
    if (a->items[i] == m)
      return i;
  }
  return -1;
}

/*+
 * bandJoinComplement - Join the complements of a pair of IntLists
 *
 * usage:
 *        bandJoinComplement(a,b)
 *
 * args:
 *        a - first int list
 *        b - second int list
 *
 * returns:
 *        - NULL -- cannot allocate response
 *        - resulting int list that contains the complement of entries 1-n
 *          where n twice the length of a and b.  I.e., n = 2*(a->n + b->n)
 *
 * notes:
 *
 * uses:
 *       bandAllocateIntList
 *       bandIntListIndexOf
 *-
 */
ae_pIntList_t
bandJoinComplement(const ae_pIntList_t a, const ae_pIntList_t b, unsigned int n)
{
  ae_pIntList_t answer = NULL;
  int i, j;

  if (!a || !b)
    return NULL;

  answer = bandAllocateIntList(n/2);
  if (answer) {
    for (i=1, j=0; i <= n/2; i++) {
      if (bandIntListIndexOf(a, i) < 0)
	answer->items[j++] = i;
    }
    for (i=(n/2)+1; i <= n; i++) {
      if (bandIntListIndexOf(b, i) < 0)
	answer->items[j++] = i;
    }
  }

  return answer;
}

/*+
 * bandCanonicalFactorToCycles - Takes a canonical factor A of B_n and computes the associated parallel
 *                               descending cycles 
 *
 * usage:
 *         bandCanonicalFactorToCycles (bl, n)
 *
 * args:
 *        bl - band list to process
 *        n - largest subscript value
 *
 * returns:
 *        - NULL - could not allocate memory
 *        - pointer to list of cycles
 *
 * notes:
 *
 * uses:
 *       bandAllocatePermutation
 *       bandMultiplyPermutations
 *       bandFreePermutation
 *       bandToCycles
 *
 *-
 */
ae_pCycles_t
bandCanonicalFactorToCycles(ae_pBandList_t bl, unsigned int n) {
  ae_pCycles_t      lp = NULL;
  ae_pPermutation_t answer,temp,twocycle = NULL;
  unsigned int      i,k;
  int               t, s;

  if (bl->n == 0) {
    answer = bandAllocatePermutation (n);
    for (i=0; i < n; i++) {
      answer->items[i] = i+1;
    }
  } else {
    t = bl->bwl[0].s[0];
    s = bl->bwl[0].s[1];
    answer = bandAllocatePermutation (n);
    for (i=0; i < n; i++) {
      if (i+1 == t) 
	answer->items[i] = s;
      else if (i+1 == s)
	answer->items[i] = t;
      else
	answer->items[i] = i+1;
    }
    k = 0;
    twocycle = bandAllocatePermutation (n);
    while (k < bl->n-1) {

      s = bl->bwl[k+1].s[0];
      t = bl->bwl[k+1].s[1];

      for (i=0; i < n; i++) {
	if (i+1 == t)
	  twocycle->items[i] = s;
	else if (i+1 == s)
	  twocycle->items[i] = t;
	else
	  twocycle->items[i] = i+1;
      }
      temp = bandMultiplyPermutations (answer,twocycle);
      bandFreePermutation (answer);
      answer = temp;
      k++;
    }
  }
  lp = bandToCycles (answer);

  bandFreePermutation (answer);
  bandFreePermutation (twocycle);

  return lp;
}
 
/*+
 * bandToCycles - gives the cycle structure of permutation p as a list of cyclic permutations
 *
 * usage:
 *         bandToCycles (p)
 *
 * args:
 *        p - permutation to convert to cyclic form
 *
 * returns:
 *        - NULL - could not allocate memory
 *        - pointer to list of cyclic permutations
 *
 * notes:
 *        Converted from the Combinatorica Mathematica library. If you have a
 *        Mathematica installation, see Packages\Combinatorica\Combinatorica.nb
 *        for more details.
 *
 * uses:
 *       bandAllocatePermutation
 *       bandAllocateCycles
 *       bandAppendToCycles
 *       bandFreeCycles
 *       bandFreePermutation
 *-
 */
ae_pListOfIntLists_t
bandToCycles (ae_pIntList_t p) {
  int k, j, first;
  ae_pIntList_t np, q, temp;
  ae_pListOfIntLists_t t, answer = NULL;
  unsigned int i;
  int ii,cases=0;

  t = bandAllocateCycles(0);
  /*
   * Make a copy of the input list
   */
  np = bandAllocatePermutation (p->n);
  for (i=0; i < p->n; i++) {
    np->items[i] = p->items[i];
  }
  q = bandAllocatePermutation (p->n);

  for (i=0; i < p->n; i++) {
    if (np->items[i] == 0) {
      bandAppendToCycles (t, bandAllocatePermutation(0));
    } else {
      j = 0;
      first = np->items[i];
      np->items[i] = 0;
      k = q->items[j++] = first;
      while (np->items[k-1] != 0) {
	q->items[j++] = np->items[k-1];
	np->items[k-1] = 0;
	k = q->items[j-1];
      }
      temp = bandAllocatePermutation (j);
      for (ii=0; ii < j; ii++) {
	temp->items[ii] = q->items[ii];
      }
      bandAppendToCycles (t, temp);
      cases++;
    }
  }
  bandFreePermutation (np);
  bandFreePermutation (q);
  /*
   * now pull out the non-null cases
   */
  answer = bandAllocateCycles (cases);
  for (i=0,k=0; i < p->n; i++) {
    if (t->p[i]->n != 0) {
      answer->p[k++] = t->p[i];
      t->p[i] = NULL;
    }
  }
  bandFreeCycles (t, 1);

  return answer;

}
 
/*+
 * bandCyclesToCanonicalFactor - converts cycles from a list of parallel descending cycles, 
 *                               into a canonical factor
 *
 * usage:
 *          bandCyclesToCanonicalFactor (l)
 *
 * args:
 *          l - cycles list
 *
 * returns:
 *          - NULL - could not allocate memory
 *          - pointer to canonical factor
 *
 * notes:
 *
 * uses:
 *       bandAllocateBandList
 *       bandFreeBandList
 *
 *-
 */
ae_pBandList_t
bandCyclesToCanonicalFactor(ae_pCycles_t l) {
  unsigned int i,j,k;
  ae_pBandList_t temp,t1,answer;

  answer = bandAllocateBandList (0);

  for (i=0; i<l->n; i++) {
    if (l->p[i]->n >1) {
      temp = bandAllocateBandList (l->p[i]->n-1);
      for (j=0; j < l->p[i]->n-1;j++) {
	temp->bwl[j].s[0] = l->p[i]->items[j];
	temp->bwl[j].s[1] = l->p[i]->items[j+1];
	temp->bwl[j].e    = 1;
      }
      /* prepend onto the answer */
      t1 = bandAllocateBandList (answer->n + temp->n);
      for (j=0; j < temp->n; j++) {
	t1->bwl[j].s[0] = temp->bwl[j].s[0];
	t1->bwl[j].s[1] = temp->bwl[j].s[1];
	t1->bwl[j].e    = temp->bwl[j].e;
      }
      for (k=0; k < answer->n; k++,j++) {
	t1->bwl[j].s[0] = answer->bwl[k].s[0];
	t1->bwl[j].s[1] = answer->bwl[k].s[1];
	t1->bwl[j].e    = answer->bwl[k].e;
      }
      bandFreeBandList (temp);
      bandFreeBandList (answer);
      answer = t1;
    }
  }
  return answer;
}
 
/*+
 * bandCyclesInverse - computes the inverse of cycles in the symmetric group
 *
 * usage:
 *        bandCyclesInverse (l)
 *
 * args:
 *       l - symmetric group to take the inverse of
 *
 * returns:
 *       - NULL - could not allocate memory
 *       - pointer to inverse cycles
 *
 * notes:
 *
 * uses:
 *       bandAllocateCycles
 *       bandAllocatePermutation
 *-
 */
ae_pCycles_t
bandCyclesInverse(ae_pCycles_t l) {
  ae_pCycles_t l1;
  ae_pPermutation_t p1;
  unsigned int *items;
  unsigned int i,k,n;

  assert (l != NULL);
  l1 = bandAllocateCycles (l->n);
  /*
   * process the list in reverse order
   */
  for (i=0; i < l->n; i++) {
    n = l->p[l->n-i-1]->n;
    items = l->p[l->n-i-1]->items;
    p1 = bandAllocatePermutation (n);
    for (k=0; k<n; k++) {
      p1->items[k] = items[n-k-1];
    }
    l1->p[i] = p1;
  }
  return l1;
}
 
/*+
 * bandMultiplyCycles - computes the product of l1 and l2 in the symmetric group
 *
 * usage:
 *        bandMultiplyCycles (l1, l2)
 *
 * args:
 *        l1 - first cycles list
 *        l2 - second cycles list
 *
 * returns:
 *        - NULL - could not allocate memory
 *        - pointer to product cycles list
 *
 * notes:
 *
 * uses:
 *       bandFromCycles
 *       bandMultiplyPermutations
 *       bandToCycles
 *       bandFreePermutation
 *
 *-
 */
ae_pCycles_t
bandMultiplyCycles(ae_pCycles_t l1, ae_pCycles_t l2) {
  ae_pPermutation_t p1,p2,p3;
  ae_pCycles_t l;

  p1 = bandFromCycles (l1);
  p2 = bandFromCycles (l2);

  p3 = bandMultiplyPermutations (p1,p2);

  l = bandToCycles (p3);

  bandFreePermutation (p1);
  bandFreePermutation (p2);
  bandFreePermutation (p3);

  return l;
}
 
/*+
 * bandCycleCompare - compare two permutation list items. This is a static
 *                    function called by qsort.
 *
 * usage:
 *          bandCycleCompare (a1, a1)
 *
 * args:
 *        a1 - pointer to first item
 *        a2 - pointer to second item
 *
 * returns:
 *        -1 - a1 < a2
 *         0 - a1 == a2
 *        +1 - a1 > a2
 *
 * notes:
 *        See man page for qsort for more details
 *
 * uses: N/A
 *
 *-
 */
static int
bandCycleCompare (const void * a1, const void * a2) {
  ae_pPermutation_t  *p1,  *p2;
  p1 = (ae_pPermutation_t *) a1;
  p2 = (ae_pPermutation_t *) a2;

  if (p1[0]->items[0] < p2[0]->items[0])
    return -1;
  if (p1[0]->items[0] > p2[0]->items[0])
    return 1;
  return 0;
}
 
/*+
 * bandFromCycles - gives the permutation that has the given cycle structure
 *
 * usage:
 *       bandFromCycles (l)
 *
 * args:
 *       l - cycles list
 *
 * returns:
 *       - NULL - could not allocate memory
 *       - pointer to permutation
 *
 * notes:
 *       this is a direct translation from the Mathematica Combinatorica package
 *
 * uses:
 *       bandAllocateCycles
 *       bandFreeCycles
 *       bandAllocatePermutation
 *       bandFreePermutation
 *       qsort - system dependent quicksort routine
 *-
 */
ae_pPermutation_t
bandFromCycles(ae_pCycles_t l) {
  ae_pCycles_t lr,f,t;
  ae_pPermutation_t p,temp;
  unsigned int i,j,k, len;
  /*
   * First step, create a copy of the Cycles, rotated one position right
   */
  lr = bandAllocateCycles (l->n);
  for (i=0; i < l->n; i++) {
    lr->p[i] = bandAllocatePermutation (l->p[i]->n);
    lr->p[i]->items[0] = l->p[i]->items[l->p[i]->n-1];
    for (k=1; k < l->p[i]->n; k++) {
      lr->p[i]->items[k] = l->p[i]->items[k-1];
    }
  }
  /*
   * create flat versions of each list
   */
  len = 0;
  for (i=0; i< l->n; i++) {
    len += l->p[i]->n;
  }
  f = bandAllocateCycles(2);
  temp = bandAllocatePermutation(len);
  k = 0;
  for (i=0; i < lr->n; i++) {   /* rotated list */
    for (j=0; j<lr->p[i]->n; j++) {
      temp->items[k++] = lr->p[i]->items[j];
    }
  }
  f->p[0] = temp;
  bandFreeCycles (lr,1);

  temp = bandAllocatePermutation(len);
  k = 0;
  for (i=0; i < l->n; i++) {   /* original list */
    for (j=0; j<l->p[i]->n; j++) {
      temp->items[k++] = l->p[i]->items[j];
    }
  }
  f->p[1] = temp;
  /*
   * next Transpose the two lists
   */
  t = bandAllocateCycles (len);
  for (i=0; i < len; i++) {
    t->p[i] = bandAllocatePermutation(2);
    t->p[i]->items[0] = f->p[0]->items[i];
    t->p[i]->items[1] = f->p[1]->items[i];
  }
  bandFreeCycles (f,1);
  /*
   * now sort the transposed list
   */
  qsort (t->p,t->n,sizeof(ae_pPermutation_t *),bandCycleCompare);
  /*
   * pull out the final permutation
   */
  p = bandAllocatePermutation (len);
  for (i=0; i < len; i++) {
    p->items[i] = t->p[i]->items[1];
  }
  bandFreeCycles (t,1);

  return p;
}

 
/*+
 *
 * bandRemoveSingletons - removes all length one cycles from a list of cycles
 *
 * usage:
 *         bandRemoveSingletons (l)
 *
 * args:
 *        l - pointer to cycle list
 *
 * returns:
 *        - NULL - could not allocate memory
 *        - pointer to resultant cycle list
 *
 * notes:
 *
 * uses:
 *        bandAllocateCycles
 *        bandAllocatePermutation
 *
 *-
 */
ae_pCycles_t
bandRemoveSingletons (ae_pCycles_t l){
  ae_pCycles_t answer;
  unsigned int i,j=0,k,len=0;
  /*
   * figure out how many cycles will remain
   */
  for (i=0; i<l->n; i++) {
    if (l->p[i]->n > 1)
      len++;
  }
  answer = bandAllocateCycles (len);
  for (i=0; i<l->n;i++) {
    if (l->p[i]->n > 1) {
      answer->p[j] = bandAllocatePermutation(l->p[i]->n);
      for (k=0; k < l->p[i]->n;k++) {
	answer->p[j]->items[k] = l->p[i]->items[k];
      }
      j++;
    }
  }
  return answer;
}
 
/*+
 *
 * bandRightComplement - computes the right complement A^* of the canonical factor A where A A^* = Delta
 *
 * usage:
 *       bandRightComplement (bl, n)
 *
 * args:
 *       bl - pointer to the canonical factor
 *       n  - max value of subscripts
 *
 * returns:
 *       - NULL - could not allocate memory
 *       - pointer to cycle list for the right complement
 *
 * notes:
 *
 * uses:
 *       bandCanonicalFactorToCycles
 *       bandCyclesInverse
 *       bandDelta
 *       bandMultiplyCycles
 *       bandCyclesToCanonicalFactor
 *       bandFreeBandList
 *
 *-
 */
ae_pBandList_t
bandRightComplement (ae_pBandList_t bl, unsigned int n) {
  ae_pCycles_t a,d,ainv,l;
  ae_pBandList_t delta, answer;

  a = bandCanonicalFactorToCycles(bl,n);
  ainv = bandCyclesInverse(a);

  delta = bandDelta(n);
  d = bandCanonicalFactorToCycles(delta,n);

  l = bandMultiplyCycles(ainv,d);
  answer = bandCyclesToCanonicalFactor (l);

  bandFreeCycles (l,1);
  bandFreeCycles (a,1);
  bandFreeCycles (d,1);
  bandFreeCycles (ainv,1);
  bandFreeBandList (delta);

  return answer;
}
 
/*+
 *
 * bandMeetOfTwoSingleDescendingCycles - omputes the meet (intersection) of two descending cycles Sigma,Tau
 *
 * usage:
 *       bandMeetOfTwoSingleDescendingCycles (sigma,tau)
 *
 * args:
 *        sigma - first cycle
 *        tau   - second cycle
 *
 * returns:
 *         - NULL - could not allocate memory
 *         - pointer to resulting cycle
 *
 * notes:
 *
 * uses:
 *       bandAllocatePermutation
 *       bandAppendToPermutation
 *       bandFreePermutation
 *
 *-
 */
ae_pPermutation_t
bandMeetOfTwoSingleDescendingCycles(ae_pPermutation_t sigma, ae_pPermutation_t tau) {
  ae_pPermutation_t xi/*,temp*/;
  unsigned int i,k;

  xi = bandAllocatePermutation (0);

  for (i=0; i<sigma->n; i++) {
    for (k=0; k < tau->n; k++) {
      if (sigma->items[i] == tau->items[k]) {
	bandAppendToPermutation (xi,sigma->items[i]);
      }
    }
  }
  /*
   * In the model, they need to reverse the list since Intersection 
   * starts at the end of the list and goes backwards. The inverse 
   * is done here, but I'm leaving the reversal code in here in 
   * case its needed someday.
   */
#if 0
  temp = bandAllocatePermutation (xi->n);
  for (i=0; i < temp->n; i++) {
    temp->items[i] = xi->items[xi->n-i-1];
  }
  bandFreePermutation (xi);
  xi = temp;
#endif
  return xi;
}
 
/*+
 *
 * bandMeetOfTwoCanonicalFactors - computes the meet  of two canonical factors A, B on 
 *                           the braid group B_n 
 *
 * usage:
 *         bandMeetOfTwoCanonicalFactors (a,b,n)
 *
 * args:
 *        a - first canonical factor
 *        b - second canonical factor
 *        n - max value of subscripts in a and b
 *
 * returns:
 *       - NULL - could not allocate memory
 *       - pointer to resultant factor
 *
 * notes:
 *
 * uses:
 *       bandCanonicalFactorToCycles
 *       bandRemoveSingletons
 *       bandFreeCycles
 *       bandAllocateCycles
 *       bandMeetOfTwoSingleDescendingCycles
 *       bandAppendToCycles
 *       bandCyclesToCanonicalFactor
 *
 *-
 */
ae_pBandList_t
bandMeetOfTwoCanonicalFactors (ae_pBandList_t a, ae_pBandList_t b, unsigned int n) {
  ae_pBandList_t answer;
  ae_pPermutation_t t;
  ae_pCycles_t cyclesA, cyclesB, temp;
  unsigned int i, j;

  temp = bandCanonicalFactorToCycles(a,n);
  cyclesA = bandRemoveSingletons(temp);
  bandFreeCycles(temp,1);

  temp = bandCanonicalFactorToCycles(b,n);
  cyclesB = bandRemoveSingletons(temp);
  bandFreeCycles(temp,1);

  temp = bandAllocateCycles (0);
  for (i=0; i < cyclesA->n; i++) {
    for (j=0; j < cyclesB->n; j++) {
      t = bandMeetOfTwoSingleDescendingCycles(cyclesA->p[i],cyclesB->p[j]);
      bandAppendToCycles (temp,t);
    }
  }
  answer = bandCyclesToCanonicalFactor(temp);

  bandFreeCycles(temp,1);
  bandFreeCycles(cyclesA,1);
  bandFreeCycles(cyclesB,1);

  return answer;
}

/*+
 *  bandGeneralRightComplement - computes the B right complement B'  where B = MB' with M = Meet
 *
 * usage:
 *        bandGeneralRightComplement (bl, meet, n)
 *
 * args:
 *        bl - band word list to find right complement of
 *        meet - meet band word list
 *        n - max subscript value
 *
 * returns:
 *        - NULL - could not allocate memory
 *        - pointer to B right complement
 *
 * notes:
 *
 * uses:
 *      bandCanonicalFactorToCycles
 *      bandCyclesInverse
 *      bandFreeCycles
 *      bandMultiplyCycles
 *      bandCyclesToCanonicalFactor
 *-
 */
ae_pBandList_t
bandGeneralRightComplement (ae_pBandList_t bl, ae_pBandList_t meet, unsigned int n) {
  ae_pCycles_t BCycle,MeetCycle, t;
  ae_pBandList_t answer;

  BCycle = bandCanonicalFactorToCycles (bl, n);
  MeetCycle = bandCanonicalFactorToCycles (meet, n);

  t = bandCyclesInverse (MeetCycle);
  bandFreeCycles (MeetCycle,1);
  MeetCycle = t;
  t = bandMultiplyCycles (MeetCycle,BCycle);
  bandFreeCycles (MeetCycle,1);
  bandFreeCycles (BCycle, 1);

  answer = bandCyclesToCanonicalFactor (t);
  bandFreeCycles (t,1);

  return answer;
}
 
/*+
 * bandLocalLeftWeight - Replaces  the product of two canonical factors AB by A' B' where A' 
 *                       B' is left weighted
 *
 * usage:
 *       bandLocalLeftWeight (A, B, n)
 *
 * args:
 *       A - first canonical factor
 *       B - second canonical factor
 *       n - max value of subscripts in A and B
 *
 * returns:
 *       - NULL - could not allocate memory
 *       - pointer to product of the two factors
 *
 * notes:
 *
 * uses:
 *      bandRightComplement
 *      bandAllocateBandList
 *      bandMeetOfTwoCanonicalFactors
 *      bandFreeBandList
 *      bandCloneBandList
 *      bandGeneralRightComplement
 *-
 */
ae_pBandList_t
bandLocalLeftWeight (ae_pBandList_t A, ae_pBandList_t B, unsigned int n) {
  ae_pBandList_t AStar,CC, BPrime, APrime, answer;
  unsigned int i,k;

  AStar = bandRightComplement (A,n);
  if (AStar->n == 0) {
    CC = bandAllocateBandList (0);
  } else {
    CC = bandMeetOfTwoCanonicalFactors (AStar, B, n);
  }
  bandFreeBandList (AStar);

  if (CC->n == 0) {
    BPrime = bandCloneBandList(B);
  } else {
    BPrime = bandGeneralRightComplement (B, CC, n);
  }
  APrime = bandAllocateBandList (A->n+CC->n);
  for (i=0; i < A->n; i++) {
    APrime->bwl[i].bl   = bandCloneBandList(A->bwl[i].bl);
    APrime->bwl[i].d    = A->bwl[i].d;
    APrime->bwl[i].s[0] = A->bwl[i].s[0];
    APrime->bwl[i].s[1] = A->bwl[i].s[1];
    APrime->bwl[i].e    = A->bwl[i].e;
  }
  for (k=0; k< CC->n; k++,i++) {
    APrime->bwl[i].bl = bandCloneBandList(CC->bwl[k].bl);
    APrime->bwl[i].d    = CC->bwl[k].d;
    APrime->bwl[i].s[0] = CC->bwl[k].s[0];
    APrime->bwl[i].s[1] = CC->bwl[k].s[1];
    APrime->bwl[i].e    = CC->bwl[k].e;
  }
  bandFreeBandList (CC);

  answer = bandAllocateBandList (2);

  answer->bwl[0].bl = APrime;
  answer->bwl[1].bl = BPrime;

  return answer;
}
 
/*+
 * bandBackwardsIteratedLeftWeight - Iterate using backwardsLeftWeight over a list of factors
 *
 * usage:
 *        bandBackwardsIteratedLeftWeight (bl, n)
 *
 * args:
 *      bl - list of factors to iterate over
 *      n  - max value of subscripts
 *
 * returns:
 *      NULL - could not allocate memory
 *      list of left weighted factors
 *
 * notes:
 *
 * uses:
 *       bandCloneBandList
 *       bandLocalLeftWeight
 *       bandFreeBandList
 *-
 */
ae_pBandList_t
bandBackwardsIteratedLeftWeight (ae_pBandList_t bl, unsigned int n) {
  ae_pBandList_t answer, LW;
  unsigned int i,j;

  answer = bandCloneBandList (bl);
  i = answer->n -2;
  j = 0;
  while (j <= i) {
    LW = bandLocalLeftWeight (answer->bwl[i-j].bl,answer->bwl[i+1-j].bl,n);
    if (answer->bwl[i-j].bl) {
      bandFreeBandList (answer->bwl[i-j].bl);
    }
    answer->bwl[i-j].bl   = bandCloneBandList (LW->bwl[0].bl);
    answer->bwl[i-j].d    = LW->bwl[0].d;
    answer->bwl[i-j].s[0] = LW->bwl[0].s[0];
    answer->bwl[i-j].s[1] = LW->bwl[0].s[1];
    answer->bwl[i-j].e    = LW->bwl[0].e;

    if (answer->bwl[i+1-j].bl) {
      bandFreeBandList (answer->bwl[i+1-j].bl);
    }
    answer->bwl[i+1-j].bl   = bandCloneBandList (LW->bwl[1].bl);
    answer->bwl[i+1-j].d    = LW->bwl[1].d;
    answer->bwl[i+1-j].s[0] = LW->bwl[1].s[0];
    answer->bwl[i+1-j].s[1] = LW->bwl[1].s[1];
    answer->bwl[i+1-j].e    = LW->bwl[1].e;
    j++;
    bandFreeBandList (LW);
  }

  return answer;
}
 
/*+
 * bandForwardsIteratedLeftWeight - iterate over a factor list in a forward direction using
 *                                 bandBackwardsIteratedLeftWeight
 *
 * usage:
 *       bandForwardsIteratedLeftWeight (bl, n, percent)
 *
 * args:
 *      bl - list of factors to iterate over
 *      n - max value of subscripts
 *      percent -- percent of bandwords to iterate backwards
 *
 * returns:
 *      NULL - could not allocate memory
 *      pointer to list of left weighted factors
 *
 * notes:
 *
 * uses:
 *       bandLocalLeftWeight
 *       bandAllocateBandList
 *       bandCloneBandList
 *       bandAppendToBandList
 *       bandBackwardsIteratedLeftWeight
 *       bandFreeBandList
 *-
 */
static ae_pBandList_t
bandForwardsIteratedLeftWeight (ae_pBandList_t bl, unsigned int n, unsigned int percent) {
  ae_pBandList_t LW,answer,temp;
  unsigned int i;

  if (percent > 0)
    percent--;

  LW = bandLocalLeftWeight (bl->bwl[0].bl, bl->bwl[1].bl, n);
  if (LW->bwl[1].bl->n == 0) {
    answer = bandAllocateBandList (1);
    answer->bwl[0].bl = bandCloneBandList (LW->bwl[0].bl);
  } else {
    answer = LW;
  }
  for (i=2; i < bl->n; i++) {
    bandAppendToBandList (answer, bl->bwl[i].bl);
    if (percent >= 99 || braidGetRandom(100) <= percent) {
      temp = bandBackwardsIteratedLeftWeight(answer,n);
      bandFreeBandList (answer);
      answer = temp;
    }
  }
  return answer;
}
 
/*+
 * bandSetOfCanonicalFactorsToBandWord - take a list of canonical factors and convert
 *                                       them to BandWords
 *
 * usage:
 *       bandSetOfCanonicalFactorsToBandWord (bl)
 *
 * args:
 *       bl - list of canonical factors
 *
 * returns:
 *       NULL - could not allocate memory
 *       pointer to list of band words
 *
 * notes:
 *
 * uses:
 *      bandCloneBandList
 *      bandAppendToBandListEx
 *-
 */
ae_pBandList_t
bandSetOfCanonicalFactorsToBandWord (ae_pBandList_t bl) {
  ae_pBandList_t answer;
  unsigned int i;

  answer = bandCloneBandList (bl->bwl[0].bl);
  if (bl->n > 1) {
    for (i=1; i < bl->n; i++) {
      bandAppendToBandListEx (answer, bl->bwl[i].bl);
    }
  }
  return answer;
}
 
/*+
 * bandRandomBandWord - create a random band word list 
 *
 * usage:
 *       bandRandomBandWord (n, l)
 *
 * args:
 *      n - maximum(ceiling) value for the subscripts
 *      l - number of bandwords to allocate
 *
 * returns:
 *         NULL - could not allocate memory
 *         ae_pBandList_t - pointer to allocated band list
 *
 * notes:
 *
 * uses:
 *      bandAllocateBandList - allocate an empty band list
 *      braidGetRandomInRange - get a random value in the specified range
 *-
 */
ae_pBandList_t
bandRandomBandWord (unsigned int n, unsigned int l) {
  ae_pBandList_t bl = NULL;
  unsigned int i;

  if (l > 0) {
    if ((bl = bandAllocateBandList (l)) != NULL) {
      for (i=0; i < l; i++) {
	bl->bwl[i].s[0] = braidGetRandom(n-1) + 2;
	bl->bwl[i].s[1] = braidGetRandom(bl->bwl[i].s[0]-1) + 1;
	bl->bwl[i].e = 2 * braidGetRandom(2) - 1;
      }
    }
  }
  return bl;
}
 
/*+
 * bandStandardCanonicalFactor - generate a standard canonical factor from
 *                a list of factors by first convert to cycles then back to factors.
 *
 * usage:
 *        bandStandardCanonicalFactor (bl, n)
 *
 * args:
 *       bl - list of factors
 *       b - max value of subscripts
 *
 * returns:
 *       NULL - could not allocate memory
 *       pointer to list of canonical factors
 *
 * notes:
 *
 * uses:
 *      bandCanonicalFactorToCycles
 *      bandCyclesToCanonicalFactor
 *      bandFreeCycles
 *-
 */
ae_pBandList_t
bandStandardCanonicalFactor(ae_pBandList_t bl, unsigned int n) {
  ae_pCycles_t c;
  ae_pBandList_t answer;

  c = bandCanonicalFactorToCycles (bl,n);
  answer = bandCyclesToCanonicalFactor (c);

  bandFreeCycles (c,1);

  return answer;
}
 
/*+
 * bandStandardSetOfCanonicalFactors - canonicalize a list of factors
 *
 * usage:
 *        bandStandardSetOfCanonicalFactors (bl, n)
 *
 * args:
 *        bl - list of factors
 *        n - max value of subscripts
 *
 * returns:
 *         NULL - could not allocate memory
 *         pointer to list of canonical factors
 *
 * notes:
 *
 * uses:
 *         bandCloneBandList
 *         bandAllocateBandList
 *         bandStandardCanonicalFactor
 *         bandFreeBandList
 *-
 */
ae_pBandList_t
bandStandardSetOfCanonicalFactors(ae_pBandList_t bl, unsigned int n) {
  ae_pBandList_t S,answer;
  unsigned int i,j;
  /*
   * get count of non-nil factors
   */
  for (i=0,j=0; i < bl->n; i++) {
    if (bl->bwl[i].bl != NULL && bl->bwl[i].bl->n)
      j++;
  }
  S = bandAllocateBandList (j);
  for (i=0,j=0; i < bl->n; i++) {
    if (bl->bwl[i].bl != NULL && bl->bwl[i].bl->n) {
      S->bwl[j++].bl = bandCloneBandList (bl->bwl[i].bl);
    }
  }
  answer = bandAllocateBandList (S->n);
  for (i=0; i < S->n; i++) {
    answer->bwl[i].bl = bandStandardCanonicalFactor(S->bwl[i].bl,n);
  }
  bandFreeBandList (S);

  return answer;
}
 
/*+
 * bandBKLNormalForm - convert a list of factors to Birman-Ko-Lee normal form
 *
 * usage:
 *       bandBKLNormalForm (bl, n)
 *       bandBKLNormalFormWithCarry (bl, n, len, delta)
 *
 * args:
 *       bl - list of factors
 *       n - max value of subscripts
 *       len - length of bandword (don't trust bl->n)
 *       delta - carryover from previous loop
 *
 * returns:
 *       NULL - could not allocate memory
 *       pointer to list of factors in BKL form
 *
 * notes:
 *
 * uses:
 *       bandFirstBKLRewrite
 *       bandAllocateBandList
 *       bandAppendToBandList
 *       bandFreeBandList
 *       bandForwardsIteratedLeftWeight
 *       bandStandardSetOfCanonicalFactors
 *       bandFreeBandList
 *       bandDelta
 *       bandFreeBandList
 *-
 */
//#define BKL_DBG
//#define BADCONJTEST_DBG
ae_pBandList_t
bandBKLNormalFormWithCarry(ae_pBandList_t bl, unsigned int n, unsigned int len, int u, unsigned int percent) {
  ae_pBandList_t bw, fw1, t, delta, answer;
  ae_pBandWord_t bwl;
  unsigned int i,k;
  int numcopies=0;

  bw = bandFirstBKLRewriteWithCarry (bl, n, len, u);
#ifdef BADCONJTEST_DBG
  // bandDumpBandList(stderr, bw); // agreement here
#endif

  // Save off delta
  u = bw->bwl[0].d;

  // Temporarily pull off the leading delta from this bandlist
  bwl = bw->bwl;
  bw->n--;
  bw->bwl = &bwl[1];

#ifdef BADCONJTEST_DBG
  // bandDumpBandList(stderr, bw); // agreement here
#endif
  t = bandForwardsIteratedLeftWeight (bw, n, percent);
#ifdef ABDCONJTEST_DBG
  // bandDumpBandList(stderr, t); // agreement here
#endif
  fw1 = bandStandardSetOfCanonicalFactors (t,n);
#ifdef BADCONJTEST_DBG
  // bandDumpBandList(stderr, fw1); // agreement here
#endif

  // Replace bw's bandlist before freeing
  bw->n++;
  bw->bwl = bwl;
  bandFreeBandList(bw);
  bandFreeBandList (t);

  delta = bandDelta (n);
#ifdef BKL_DBG
  bandPrintBandList(stderr, delta, "delta", 80);
#endif
  /*
   * search for delta in fw1, if found, update u and delete the copy
   */
  answer =bandAllocateBandList (1);
  answer->bwl[0].d = u;
#ifdef BKL_DBG
  fprintf(stderr,"{");
#endif
  for (i=0; i < fw1->n; i++) {
#ifdef BKL_DBG
    //fprintf(stderr, "  \n(%d) ", i+1);
    //bandPrintBandList(stderr, fw1->bwl[i].bl, "fw1->bwl[i].bl", 80);
#endif
    if (fw1->bwl[i].bl->n == delta->n) {
      for (k=0; k < delta->n; k++) {
	if (fw1->bwl[i].bl->bwl[k].s[0] != delta->bwl[k].s[0] &&
	    fw1->bwl[i].bl->bwl[k].e != delta->bwl[k].e)
	  break;
      }
      if (k == delta->n) {
	u++;
#ifdef BKL_DBG
	if (numcopies)
	  fprintf(stderr,", ");
	fprintf(stderr,"%d",i+1);
#endif
	numcopies++;
      }
    }
    else {
      bandAppendToBandList (answer, fw1->bwl[i].bl);
    }
  }

#ifdef BKL_DBG
  fprintf(stderr, "}\n");
  fprintf(stderr, "Found %d copies of Delta\n",numcopies);
#endif
  answer->bwl[0].d = u;


  bandFreeBandList (fw1);
  bandFreeBandList (delta);

  /*
    if (u == 0) {
    bandFreeBandList (answer);
    return NULL; 
    }
  */
  
#ifdef BADCONJTEST_DBG
  bandDumpBandList(stderr, answer); // and they agree here!
#endif

  return answer;
}
/*+
 * bandRandomSampleOfIntList - builds an integer list by randomly sampling from a given one.
 *
 *
 * usage:
 *        bandRandomSampleOfIntList(s, l)
 *
 * args:
 *       s - integer list to sample from
 *       l - number to take in random sample
 *
 * returns: NULL - could not allocate
 *          ae_pIntList_t - pointer to allocated integer list
 *
 * notes:
 *       
 *
 * uses:
 *         bandCloneIntList
 *         bandAllocateIntList
 *         bandRandomPermutation
 *         bandFreePermutation
 *-
 */
//#define RSIL_DBG
ae_pIntList_t
bandRandomSampleOfIntList(ae_pIntList_t s, int l) 
{
  ae_pIntList_t result = NULL;
  ae_pPermutation_t indexlist = NULL;
  int original_length = s->n;
  int i = 0;

  if (l >= original_length) {
    result = bandCloneIntList(s);
    return result;
  }

  if (l <= 0) {
    result = bandAllocateIntList(0);
    return result;
  }
  
  result = bandAllocateIntList(l);
  if (!result)
    return NULL;


  indexlist = bandRandomPermutation(original_length);

#ifdef RSIL_DBG
  bandPrintPermutationFlat(stdout, indexlist);
#endif


  for (i=0; i < l; i++) {
    result->items[i] = s->items[ indexlist->items[i]-1 ];
  }

  bandFreePermutation(indexlist);

  return result;
}

ae_pBandList_t
bandBKLNormalForm(ae_pBandList_t bl, unsigned int n) {
  unsigned int len;
  int delta = 0;

  if (!bl)
    return NULL;

  // Determine the length and delta-carry
  len = bl->n;
  if (bl->bwl[len-1].d) {
    len--;
    delta = bl->bwl[len].d;
  }

  return bandBKLNormalFormWithCarry(bl, n, len, delta, 100);
}

/*+
 *
 * bandRandomPermutation - generate a random permutation based on n
 *
 * usage: bandRandomPermutation(n)
 *
 * args:
 *       n - length and max value of permutation items
 *
 * returns:
 *      pointer to permutation array
 *
 * notes:
 *      return pointer must be freed by caller
 *
 * uses:
 *      bandAllocatePermutation
 *      braidGetRandomInRange
 *
 *-
 */

/* Whether or not we use the classic Fisher Yates
 * algorithm to generate permuations.
 * 
 * See http://en.wikipedia.org/wiki/Fisher-Yates_shuffle
 */
#define FISHER_YATES_ALGORITHM 1
ae_pPermutation_t
bandRandomPermutation (unsigned int  n) {
  int i,j;
  ae_pPermutation_t p;
  int tmp;
#ifndef FISHER_YATES_ALGORITHM
  int r,k,found;
#endif
  
  p = bandAllocatePermutation (n);
     
#ifdef FISHER_YATES_ALGORITHM
  for (i=0; i<n; i++) 
    p->items[i] = i+1;
  for (i=n-1; i !=0 ; i--) {
    j = braidGetRandom(i+1);
    tmp = p->items[i];
    p->items[i]=p->items[j];
    p->items[j]=tmp;
  }
  return p;
#else
  /* original implementation */
  for (i=0; i < n; i++) {
    while (1) {
      r =  braidGetRandomInRange (1, n);
      if (i == 0)
	break;
      found = 0;
      for (k=0; k <i; k++) {
	if (r == p->items[k]) {
	  found = 1;
	  break;
	}
      }
      if (!found)
	break;
    }
    p->items[i] = r;
  }
  return p;
#endif
}
 
/*+
 * bandDrop - drop an element of a band list
 *
 * usage:  bandDrop (bl, elem)
 *
 * args:
 *        bl   - bandlist to operate on
 *        elem - element number to drop
 *
 * returns:
 *        new bandlist with the element removed
 *
 * notes:
 *
 * uses:
 *
 *-
 */
ae_pBandList_t
bandDrop (ae_pBandList_t bl, unsigned int elem) {
  ae_pBandList_t p = NULL;
  unsigned int i,k;

  if (bl) {
    p = (ae_pBandList_t) bandAllocateBandList (bl->n-1);
    if (p) {
      if (bl->n > 0) {
	for (i=0,k=0; i < bl->n; i++) {
	  if (elem == i) /* if element to be removed, skip copying */
	    continue;
	  /* sublist entry */
	  if (bl->bwl[i].bl) {
	    p->bwl[k++].bl = bandCloneBandList (bl->bwl[i].bl);
	  } else if (bl->bwl[i].d) {  /* delta */
	    p->bwl[k++].d = bl->bwl[i].d;
	  } else {                    /* bandword */
	    p->bwl[k].s[0] = bl->bwl[i].s[0];
	    p->bwl[k].s[1] = bl->bwl[i].s[1];
	    p->bwl[k++].e    = bl->bwl[i].e;
	  }
	}
      } else {
	p->n = 0;
	p->bwl = NULL;
      }
    }
  }
  return p;
}
 
/*+
 * bandAllocateTranspositions - allocate a transposition list
 *
 * usage: bandAllocateTranspositions (len)
 *
 * args:
 *      len - length of transposition to allocate
 *
 * returns:
 *      pointer to transposition list or NULL if memory could not be allocated
 *
 * notes:
 *
 * uses:
 *      ae_calloc
 *      ae_free
 *-
 */ 
ae_pTranspositions_t
bandAllocateTranspositions (unsigned int len) {
  ae_pTranspositions_t p;

  p = (ae_pTranspositions_t) ae_calloc (1, sizeof (ae_Transpositions_t));
  if (!p)
    return NULL;

  p->len = len;
  p->v = (ae_pTransposition_t *) ae_calloc (len, sizeof (ae_pTransposition_t));
  if (!p->v) {
    ae_free (p);
    return NULL;
  }
  return p;
}

/*+
 * bandAllocateTransposition - allocate a transposition
 *
 * usage: bandAllocateTransposition (len)
 *
 * args:
 *      len - length of transposition to allocate
 *
 * returns:
 *      pointer to transposition or NULL if memory could not be allocated
 *
 * notes:
 *
 * uses:
 *      ae_calloc
 *      ae_free
 *-
 */
ae_pTransposition_t
bandAllocateTransposition (unsigned int len) {
  ae_pTransposition_t p;

  p = (ae_pTransposition_t) ae_calloc (1,sizeof(ae_Transposition_t));
  if (!p)
    return NULL;
  p->len = len;
  p->data = (ae_pTranspositionData_t) ae_calloc (len, sizeof(ae_TranspositionData_t));
  if (!p->data) {
    ae_free (p);
    return NULL;
  }
  return p;
}
void
bandPrintTranspositions	(FILE * fd, ae_pTranspositions_t p, char * comment)
{
  int i;
  fprintf(stderr,"%s ",comment);
  for (i=0; i<p->len; i++) {
    fprintf(fd,"(%d,%d)",p->v[i]->data->l,p->v[i]->data->r);
  }
  fprintf(fd,"\n");
}
/*+
 * bandFreeTransposition - free a previously allocated transposition
 *
 * usage: bandFreeTransposition (ptr)
 *
 * args:
 *      ptr - pointer to transposition to free
 *
 * returns: N/A
 *
 * notes:
 *
 * uses:
 *       ae_free
 *
 *-
 */
void
bandFreeTransposition (ae_pTransposition_t p) {
  if (p) {
    if (p->data)
      ae_free (p->data);
    ae_free (p);
  }
}

/*+
 * bandFreeTranspositions - free a list of transpositions
 *
 * usage: bandFreeTransposition (ptr, flags)
 *
 * args:
 *      ptr - pointer to list of transpositions
 *      flags - if 1, then free all transpositions as well
 *
 * returns: N/A
 *
 * notes:
 *
 * uses:
 *      ae_free
 *
 *-
 */
void
bandFreeTranspositions (ae_pTranspositions_t p, unsigned int flags) {
  if (p) {
    if (p->v) {
      if (flags) {
	int i;
	for (i=0; i<p->len; i++) {
	  bandFreeTransposition(p->v[i]);
	  p->v[i] = NULL;
	}
      }
      ae_free (p->v);
    }
    ae_free (p);
  }
}

/*+
 * bandAppendTranspositionsToTranspositions - concatenate two transpositions
 *
 * usage: bandAppendTranspositionsToTranspositions (d,s)
 *
 * args:
 *      d - destination transposition
 *      a - source transposition
 *
 * returns: N/A
 *
 * notes:
 *
 * uses:
 *     ae_realloc
 *
 *-
 */
void
bandAppendTranspositionsToTranspositions (ae_pTranspositions_t d, ae_pTranspositions_t a) {
  int i;
  d->v = (ae_pTransposition_t *) ae_realloc (d->v,(d->len+a->len)*sizeof(ae_pTransposition_t));
  for (i=0; i < a->len; i++) {
    d->v[d->len++] = a->v[i];
  }
}


/* \delta\tau(n) = n\delta.  Used to move rightmost deltas to the
 * beginning of the permutation list
 *
 * As defined in the paper,
 * \tau(a_{ts}) = a_{(t+1)(s+1) and
 * \tau^{-1}(a_{ts}) = a_{(t-1)(s-1).
 *
 * delta refers to the number of iterations which tau must run.  That
 * is, to move \delta^{n} to the left of the permutation perm, apply
 * \tau^{n}(perm)
 */
static void tau(uint8_t *perm, int delta, void *tmp, uint8_t index)
{
	uint8_t i;
	int *difference = tmp;

	delta = delta % index;

	if(delta < 0) {
		delta += index;
	}

	for(i=0; i<index; i++) {
		difference[i] = (perm[i] - i) % index;
	}

	for(i=0; i<index; i++) {
		int idx = (i - delta) % index;
		int t;

		if(idx < 0) {
			idx += index;
		}

		t = (i + difference[idx]) % index;
		if(t < 0) {
			t += index;
		}

		perm[i] = (uint8_t) t;
	}
}

enum {
	SIGN_BIT = 0x80
};

/* multiply band generator that takes `to' to `from' by a permutation
 * perm.
 */
static void band_mul(uint8_t to, uint8_t from, uint8_t *perm,
		uint8_t index)
{
	uint8_t toaddr = 0;
	uint8_t fromaddr = 0;
	uint8_t i;
	for(i=0; i<index; i++) {
		if(perm[i] == to) {
			toaddr = i;
		}

		if(perm[i] == from) {
			fromaddr = i;
		}
	}

	perm[fromaddr] = to;
	perm[toaddr] = from;
}

/* Convert as many Artin generators as possible to the permutation
 * representation of a canonical factor.  This includes converting
 * negative generators to positive, descending cycles by multiplying
 * them by \delta.  This function uses a greedy algorithm to combine as
 * many positive Artin generators into a single canonical factor as
 * possible.  It will add Artin generators to a canonical factor until
 * two elements in the canonical factor form an obstructing pair, or in
 * this case if, for two positive generators a_i and a_j,
 * a_i - a_j <= 1 and
 * a_i < a_j
 *
 * TODO: it may be possible to also combine negative factors into a
 * single canonical factor and a positive delta.
 */
static unsigned int artin2canonical(uint8_t *canonical, const uint8_t *artin,
  unsigned int count, const uint8_t *identity, const uint8_t *delta,
  uint8_t index)
{
	if(*artin & SIGN_BIT) { /* negative */
		uint8_t from = *artin & ~SIGN_BIT;
		uint8_t to = (from + 1) % index;
		int t;

		memcpy(canonical, delta, index);
		t = (to - 1) % index;
		if(t < 0) {
			t += index;
		}
		canonical[from] = t;

		t = (from - 1) % index;
		if(t < 0) {
			t += index;
		}
		canonical[to] = t;


		return 1;
	} else {
		unsigned int i;
		int last = -1;
		memcpy(canonical, identity, index);

		for(i=0; i<count; i++) {
			uint8_t to = *artin;
			uint8_t from = (to + 1) % index;
			bool isnegative = (*artin & SIGN_BIT) != 0;

			/* combine only positive, descending cycles */
			if(isnegative || to <= last) {
				break;
			}

			/* this heuristic algorithm won't necessarily
			 * work if the band wraps from index to 0, so
			 * don't bother combining those elements
			 */
			if(last != -1 && from == 0) {
				break;
			}

			last = to;
			band_mul(to, from, canonical, index);

			artin--;
		}

		return i;
	}
}

/* Apply steps 1 and 2 in the bkl paper by iterating backwards over the
 * array of Artin generators pointed to by artin once.  That is,
 * delta^{c->delta} is at index i and all canonical factors to the right
 * of i are positive and contain no deltas.
 */
static unsigned int first_rewrite(bklctx *c, const uint8_t *artin,
  unsigned int count)
{
	unsigned int i;
	unsigned int processed = 0;
	unsigned int canonical_count = 0;
	uint8_t *perm =  c->permutation;

	/* iterate backwards through count generators in artin,
	 * converting each generator to a permutation
	 */
	for(i=0; i < count; i += processed, canonical_count++) {
		uint8_t *dest = &perm[(c->permsize - 1 -
		  canonical_count) * c->index];

		processed = artin2canonical(dest,
		  &artin[count - i - 1], count - i,
		  c->identity_perm, c->delta_perm, c->index);
		tau(dest, c->delta, c->temp, c->index);

		/* if the last canonical factor was negative
		 * artin2canonical multiplied it by \delta.  Prepend
		 * \delta^{-1}
		 */
		if(artin[count - i-1] & SIGN_BIT) {
			c->delta--;
		}
	}

	/* Some Artin generators may not have been combined so the array
	 * of permutations may not be full.  Move the permutations back
	 * to the beginning of the struct.
	 */
	memmove(perm, &perm[(c->permsize -
	  canonical_count) * c->index], canonical_count * c->index);
	return canonical_count;
}

static const uint8_t logtable[256] = {
#define LT(n) n, n, n, n, n, n, n, n, n, n, n, n, n, n, n, n
	0, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3,
	LT(4), LT(5), LT(5), LT(6), LT(6), LT(6), LT(6),
	LT(7), LT(7), LT(7), LT(7), LT(7), LT(7), LT(7), LT(7)
};


static uint8_t high_bit(uint64_t n)
{
	uint8_t ret = 0x0;
	for(; n > 0xFF; n >>= 8) {
		ret += 8;
	}

	ret += logtable[n];
	return ret;
}

/* computes the meet (denoted by ^ in the paper) of a and b.
 * Informally, the meet of two canonical factors is the union of the
 * cycles in both a and b.  Using the example in the paper,
 * a = (5, 4, 1)(3,2)
 * b = (4, 2, 1)
 * a ^ b = (4, 1)
 *
 * This code creates two arrays of bitfields that denote descending
 * cycles.  Any bit set in the bitfield means that the corresponding
 * element is in the cycle.  For example, the cycles above would be
 * denoted (omitting elements with 1 cycle as above)
 * a = (0x31)(0x0C)
 * b = (0x13)
 * a & b = (0x11) = (0x11)
 * Once bitfields have been generated the meet of two elements can be
 * detected by anding values at the same index.  If a lower bit is set
 * that is the next value in the cycle.  If not, the cycle repeats and
 * the highest bit set is the beginning of the cycle.
 */
static void meet(uint8_t *out, const uint8_t *a, const uint8_t *b,
		uint64_t *temp, uint8_t index)
{

	uint64_t *a_cycles = temp;
	uint64_t *b_cycles = temp + index;

	uint8_t i;
	uint64_t mask;
	uint64_t mask_init = UINT64_MAX >> (1 + 64 - index);

	memset(a_cycles, 0, index * sizeof(a_cycles[0]));
	memset(b_cycles, 0, index * sizeof(b_cycles[0]));

	for(i=index; i>0; i--) {
		uint64_t cycle_bit = (uint64_t) 1u << (i-1);

		uint64_t acycle = a_cycles[i-1] | cycle_bit;
		a_cycles[a[i-1]] = acycle;
	}

	for(i=index; i>0; i--) {
		a_cycles[a[i-1]] = a_cycles[i-1];
	}

	for(i=index; i>0; i--) {
		uint64_t cycle_bit = (uint64_t) 1u << (i-1);

		uint64_t bcycle = b_cycles[i-1] | cycle_bit;
		b_cycles[b[i-1]] = bcycle;
	}

	for(i=index; i>0; i--) {
		b_cycles[b[i-1]] = b_cycles[i-1];
	}

	for(i=index, mask = mask_init; i>0; i--, mask >>= 1) {
		uint64_t meet = a_cycles[i-1] & b_cycles[i-1];
		uint64_t lower_cycles = meet & mask;

		if(lower_cycles) {
			out[i-1] = high_bit(lower_cycles);
		} else {
			out[i-1] = high_bit(meet);
		}
	}
}

/* mul: multiply two permutations a and b */
static void mul(uint8_t *out, const uint8_t *a, const uint8_t *b,
		uint8_t index)
{
	uint8_t i;
	for(i=0; i<index; i++) {
		out[i] = a[b[i]];
	}
}

/* find the right complement of a relative to b.  That is,
 * a * out = b
 */
static void right_complement(uint8_t *out, const uint8_t *a,
		const uint8_t *b, uint8_t *temp, uint8_t index)
{
	uint8_t i;

	for(i=0; i<index; i++) {
		temp[a[i]] = i;
	}

	mul(out, temp, b, index);
}

/* Apply the \Gamma operator to a and b, storing the resulting a' and b'
 * back in a and b.  This closely follows the algorithm described in
 * lemma 4.3.
 */
static void left_weight(uint8_t *a, uint8_t *b, uint8_t *temp,
		const uint8_t *delta, uint8_t index)
{
	uint8_t *c = temp;
	uint8_t *t0 = c + index;
	uint8_t *t1 = t0 + index;
	uint64_t *t1_meet = (uint64_t *) t1;
	uint8_t *t2 = t1 + index;

	right_complement(t0, a, delta, t2, index);
	meet(c, t0, b, t1_meet, index);

	mul(t1, a, c, index);
	right_complement(t0, c, b, t2, index);

	memcpy(a, t1, index);
	memcpy(b, t0, index);
}

/* unbiased check for probability */
static int roll_success(int chance)
{
	ae_uint32_t n;
	if(chance >= 100) {
		return 1;
	}

	do {
		n = braidGetRandom(1024);
	} while(n >= 1000);

	return n < (chance * 10);

}

/* Iterate backwards through the permutation array in c, left weighting
 * each value.  Each iteration may produce a delta at the beginning or
 * an identity matrix at the end.  Before each iteration, remove any
 * identity matrices at the end and add any deltas into the power of
 * delta variable c->delta.
 */
static void full_left_weight(bklctx *c, int backchance, 
		int forwardchance, int startindex)
{
	size_t i, j;
	size_t factors_begin = 0;
	size_t factors_end = 1;
	uint8_t *perm = c->permutation;

	if(startindex >= c->count) {
		startindex = c->count - 1;
	}

	factors_end += startindex;

	for(i=factors_end; i<c->count; i++) {
		memmove(&perm[factors_end * c->index],
		  &perm[i * c->index], c->index);

		if(!roll_success(forwardchance)) {
			factors_end++;
			continue;
		}

		for(j=factors_end; j>factors_begin; j--) {
			if(roll_success(backchance)) {
			left_weight(&perm[(j-1) * c->index], &perm[j * c->index], c->temp, c->delta_perm, c->index);
			}
		}

		if(!memcmp(&perm[factors_begin * c->index],
		  c->delta_perm, c->index) && factors_begin < i) {
			factors_begin++;
		}

		if(!memcmp(&perm[factors_end * c->index],
		  c->identity_perm, c->index) &&
		  factors_end > factors_begin) {
			factors_end--;
		}

		factors_end++;
	}

	c->count = factors_end - factors_begin;
	memmove(perm, &perm[factors_begin * c->index],
	  c->index * c->count);
	c->delta += factors_begin;
}

static void init_identity(uint8_t *perm, uint8_t index)
{
	uint8_t i;
	for(i=0; i<index; i++) {
		perm[i] = i;
	}
}

static void init_delta(uint8_t *perm, uint8_t index)
{
	uint8_t i;
	perm[0] = index - 1;
	for(i=1; i<index; i++) {
		perm[i] = i-1;
	}
}

static int bkl(bklctx *c, const uint8_t *artin, unsigned int count,
	int delta, uint8_t index, int backchance, int forwardchance, unsigned int startindex)
{
	c->delta = delta;
	c->permsize = count;
	c->index = index;
	init_identity(c->identity_perm, index);
	init_delta(c->delta_perm, index);
	c->count = first_rewrite(c, artin, count);

	full_left_weight(c, backchance, forwardchance, startindex);

	return c->delta;
}

ae_pBandList_t
bklwrap(ae_plist_t braid, unsigned int n, unsigned int len, int u,
        unsigned int backchance, unsigned int forwardchance, unsigned int
	startindex)

{
	bklctx c;
	uint8_t *artin;
	size_t alloc_len = n * len;
	int i;
	ae_pBandList_t ret;
	ae_pIntList_t perm;

	if(n > 64) {
		return NULL;
	}

	artin = malloc(len);

	/* allocate all bklctx data with one malloc */
	alloc_len += n * (16 + 2);
	alloc_len += n * 2;

	c.permutation = malloc(alloc_len);
	c.temp = c.permutation + (n * len);
	c.identity_perm = c.temp + (n * (16 + 2));
	c.delta_perm = c.identity_perm + n;

	if(artin == NULL || c.permutation == NULL) {
		free(artin);
		free(c.permutation);
		return NULL;
	}

	/* convert sdk braid list to internal format */
	for(i=0; i<len; i++) {
		artin[i] = braid->b[i].l - 1;

		if(braid->b[i].u == -1) {
			artin[i] |= SIGN_BIT;
		}
	}

	bkl(&c, artin, len, u, n, backchance, forwardchance, startindex);

	/* prepend band list with delta */
	ret = bandAllocateBandList (1);
	ret->bwl[0].d = c.delta;

	/* to return band list, convert output bkl
	 * permutations -> cycles -> bands
	 */
	perm = bandAllocatePermutation(n);
	for(i=0; i<c.count; i++) {
		ae_pListOfIntLists_t cycle;
		ae_pBandList_t band;
		unsigned int j;

		uint8_t *p = c.permutation;
		p += n * i;

		for(j=0; j<n; j++) {
			perm->items[j] = p[j] + 1;
		}

		cycle = bandToCycles(perm);
		band = bandCyclesToCanonicalFactor(cycle);
		bandAppendToBandList (ret, band);

		bandFreeBandList(band);
		bandFreeCycles(cycle, 1);
	}

	free(c.permutation);
	free(artin);
	bandFreePermutation(perm);

	return ret;
}

ae_pBandList_t
bandFastBKLNormalFormWithCarry(ae_plist_t braid, unsigned int n, unsigned int len, int u)
{
  return bklwrap(braid, n, len, u, 100, 100, 0);
}

ae_pBandList_t
bandFastBKLNormalForm(ae_plist_t braid, int n)
{
  return bandFastBKLNormalFormWithCarry(braid, n, braid->len, 0);
}

ae_pBandList_t
bandFastBKLNormalFormWithCarryRandomSkip(ae_plist_t braid, unsigned int n, unsigned int len, int u, unsigned int backchance, unsigned int forwardchance, unsigned int startindex)
{
  return bklwrap(braid, n, len, u, backchance, forwardchance, startindex);
}

