/*
* Copyright (c) 2017 SecureRF Corporation, All rights reserved.
* -------------------------------------------------------------------------
* This computer program is proprietary information of SecureRF Corporation.
* The receipt or possession of this program does not convey any rights to
* reproduce or disclose its contents, or to manufacture, use, or sell
* anything that it may describe, in whole or in part, except for the purposes of 
* the post-quantum algorithm public review and evaluation in connection with 
* the National Institute of Standards and Technology’s development of standards 
* for post-quantum cryptography.  Any reproduction of this program for any other 
* purpose without the express written consent of SecureRF is a violation of the 
* copyright laws and may subject you to criminal prosecution.
* -------------------------------------------------------------------------
*/


/* ae_crypto.c - Algebraic Eraser Diffie-Hellman and 
 *               Algebraic Eraser Digital Signature Algorithm 
 */
#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>
#include <string.h>
#include "ae_crypto.h"
#include <time.h>
#include "aelib.h"
#include "braids.h"
#include <nmmintrin.h>
#include <immintrin.h>
#include <arpa/inet.h>
#include <math.h>

#ifdef KAT_INTERMEDIATES
#include "printUtils.h"
#endif

#define PASTE_INTERNAL(f,b) f ## b
#define PASTE(f,b) PASTE_INTERNAL(f,b)

#ifdef __C51__
#	define	AE_LOOKUP code
#else
#	define AE_LOOKUP 
#endif

__m256i q;
__m128i q128;

int assign = 0;

#if WALNUT_SL == 128
#	define DSA_GET_LOOKUP(val) (val)
#	define DSA_ADD(valA, valB) (dsa_add(valA, valB))
#	define DSA_MUL(valA, valB) (dsa_mul(valA, valB))
#	define DSA_INVERSE(val) (dsa_inverse(val))
#	define DSA_MUL_NO_SHIFT(valA, valB) (dsa_mul_no_shift(valA, valB))

// equivalent to (a + b) % M31
uint32_t dsa_add(uint32_t a, uint32_t b)  {
	a+=b;

	if(a < WALNUT_FIELD) {
		return a;
	}
	return a - WALNUT_FIELD;
}

// equivalent to (a * b) % M31
// see https://koclab.cs.ucsb.edu/teaching/ccs130h/projects/04-finite-fields/ecc99-paar.pdf page 13
uint32_t dsa_mul(uint32_t a, uint32_t b)  {
	uint32_t hi, lo;
	uint64_t prod;

	a <<= 1;

	prod = (uint64_t)a * (uint64_t)b;

	hi = prod >> 32;
	lo = prod;

	lo >>= 1;

	return DSA_ADD(hi, lo);
}

uint32_t dsa_mul_no_shift(uint32_t a, uint32_t b)  {
	uint32_t hi, lo;
	uint64_t prod;

	prod = (uint64_t)a * (uint64_t)b;

	hi = prod >> 32;
	lo = prod;

	lo >>= 1;

	return DSA_ADD(hi, lo);
}

// extended euclidean algorithm from keygen repo
uint32_t dsa_inverse(uint32_t a) {
	int64_t u, v, x1, x2;
	uint32_t ainv;

	u = a;
	v = WALNUT_FIELD;
	x1 = 1;
	x2 = 0;

	while(u != 1 && v != 1) {

		while(!(u & 1)) {
			u >>= 1;

			if(!(x1 & 1)) {
				x1 >>= 1;
			} else {
				x1 = (x1 + WALNUT_FIELD) >> 1;
			}
		}

		while(!(v & 1)) {
			v >>= 1;

			if(!(x2 & 1)) {
				x2 >>= 1;
			} else {
				x2 = (x2 + WALNUT_FIELD) >> 1;
			}
		}

		if((u-v) >= 0) {
			u = u - v;
			x1 = x1 - x2;
		} else {
			v = v-u;
			x2 = x2-x1;
		}
	}

	if(u == 1) {
		int64_t x1mp = x1 % WALNUT_FIELD;
		ainv = x1mp < 0 ? (uint32_t)(x1mp + WALNUT_FIELD) : (uint32_t)x1mp;
	} else {
		int64_t x2mp = x2 % WALNUT_FIELD;
		ainv = x2mp < 0 ? (uint32_t)(x2mp + WALNUT_FIELD) : (uint32_t)x2mp;
	}

	return ainv;
}

#else
#       define DSA_ADD(valA, valB) (g_ffield->add(g_ffield, valA, valB))
#       define DSA_MUL(valA, valB) (g_ffield->mult(g_ffield, valA, valB))
#       define DSA_INVERSE(val) (g_ffield->inverse(g_ffield, val))
#		define DSA_MUL_NO_SHIFT(valA, valB) (g_ffield->mult_no_shift(g_ffield, valA, valB))
#endif


static uint8_t extract_elem(const uint8_t *in, size_t curr_bit, uint8_t packwidth, uint8_t mask)
{
	uint8_t ret = 0;
	int8_t shiftby = (8 - packwidth) - (curr_bit % 8);
	size_t addr = curr_bit / 8;

	if(shiftby < 0) {
		uint8_t entry1 = in[addr];
		uint8_t entry2 = in[addr + 1];
		uint8_t t = 8 + shiftby;

		ret = (entry1 << -shiftby);
		ret |= (entry2 >> t) & mask;
	} else {
		uint8_t entry = in[addr];
		ret = (entry >> shiftby);
	}

	return ret & mask;
}


static void mmul(ffelem d[WALNUT_BRAID][WALNUT_MATRIX_ROW],
		ffelem a[WALNUT_BRAID][WALNUT_MATRIX_ROW],
		ffelem b[WALNUT_BRAID][WALNUT_MATRIX_ROW])

{
	uint8_t i, j, k;

	for(i=0; i<WALNUT_BRAID; i++) {
		for(j=0; j<WALNUT_BRAID; j++) {
			d[j][i] = 0;
			for(k=0; k<WALNUT_BRAID; k++) {
				d[j][i] = DSA_ADD((d[j][i]), (DSA_MUL(a[k][i], b[j][k])));
			}
		}
	}
}

void avxAdd256(uint64_t* a, const uint64_t* b, const int32_t xor, __m256i q2) {

    //Load in 8 elements from our two arrays
    __m256i a2 = _mm256_loadu_si256((__m256i*)&a[0]);
    __m256i b2 = _mm256_loadu_si256((__m256i*)&b[0]);

    //XOR by the finite field
    if(xor == 1)
        b2 = _mm256_xor_si256 (b2, q2);

    //Add together the two arrays
    __m256i c2 = _mm256_add_epi64(a2,b2);
    __m256i test1 = _mm256_cmpeq_epi64(c2, q2);
    __m256i test2 = _mm256_cmpgt_epi64(c2, q2);
    __m256i test3 = _mm256_add_epi64(test1, test2);

//    if(c2[0] == WALNUT_FIELD || c2[1] == WALNUT_FIELD || c2[2] == WALNUT_FIELD || c2[3] == WALNUT_FIELD)
//    	printf("Test");

    //Do what is effectively a greater than or equal to operation to see which of the elements are >= WALNUT_FIELD
//    __m256i cmp = _mm256_cmpeq_epi64(test3, c2);

    //AND with the finite field to a vector containing WALNUT_FIELD elements where they need to be subtracted from
    __m256i sand = _mm256_and_si256(q2,test3);

    //Perform said subtraction
    c2 = _mm256_sub_epi64(c2, sand);

    //Store back in memory
    _mm256_storeu_si256((__m256i*)&a[0], c2);
}

void avxAdd128(uint64_t* a, const uint64_t* b, const int32_t xor, __m128i q2) {

    __m128i a2 = _mm_loadu_si128((__m128i*)&a[0]);
    __m128i b2 = _mm_loadu_si128((__m128i*)&b[0]);

    if(xor == 1)
        b2 = _mm_xor_si128 (b2, q2);

	__m128i c2 = _mm_add_epi64(a2,b2);
	__m128i test1 = _mm_cmpeq_epi64(c2, q2);
	__m128i test2 = _mm_cmpgt_epi64(c2, q2);
	__m128i test3 = _mm_add_epi64(test1, test2);

	__m128i sand = _mm_and_si128(q2,test3);

    c2 = _mm_sub_epi64(c2, sand);

    _mm_storeu_si128((__m128i*)&a[0], c2);
}


static int walnut_emul(walnut_verify_ctx *c, 
		ffelem m[WALNUT_BRAID][WALNUT_MATRIX_ROW],
		const uint8_t *b, size_t count, uint8_t config)
{
	size_t i;
	// config[7:3] determine if the braid is packed
	// config[2:0] if packed, offset to first braid bit
	uint8_t offset = config & 0x07;
	config = config & 0xF8;

	for(i=0; i<count; i++) {
		// if the braid is packed, run extract_elem
		uint8_t br = config ? extract_elem(b, i * (WALNUT_BRAID_BIT + 1) + offset,
					WALNUT_BRAID_BIT + 1, 0xFF >> (8 - (WALNUT_BRAID_BIT + 1))) : b[i];
		uint8_t r = (br & 0x0F);
		ffelem mp;
		ffelem *r0, *r1, *r2;
		uint8_t j, t0, t1;

		if(r >= WALNUT_BRAID-1) {
			break;
		}

		t0 = c->p[r]; 
		t1 = c->p[r + 1];
		
		if (t0 > WALNUT_BRAID || t1 > WALNUT_BRAID)
			return -2;

		c->p[r] = t1;
		c->p[r + 1] = t0;

		r1 = m[r];

		if (br & 0x10) {	// - sign
			r0 = r1 - WALNUT_MATRIX_ROW;
			r2 = r1 + WALNUT_MATRIX_ROW;
			mp = c->imp[t1];
		} else {				// + sign
			r0 = r1 + WALNUT_MATRIX_ROW;
			r2 = r1 - WALNUT_MATRIX_ROW;
			mp = c->mp[t0];
		}

//#define AVX

#if WALNUT_SL == 256
int iterator = 4;
int num = 8;

if(assign == 0)
{
	q = _mm256_set1_epi64x(WALNUT_FIELD);
	q128 = _mm_set1_epi64x(WALNUT_FIELD);
	assign = 1;
}
#else
int iterator = 1;
int num = WALNUT_BRAID;
#endif


		for(j=0; j<num; j+=iterator) {
#if WALNUT_SL == 256
			avxAdd256(&r0[j], &r1[j], 0, q);
#else
			r0[j] = DSA_ADD((r0[j]), (r1[j]));
#endif
		}

#if WALNUT_SL == 256
		avxAdd128(&r0[j], &r1[j], 0, q128);
#endif
		for(j=0; j<WALNUT_BRAID; j++) {
			r1[j] = DSA_MUL_NO_SHIFT(mp, r1[j]);
		}

		for(j=0; j<num; j+=iterator) {
#if WALNUT_SL == 256
			avxAdd256(&r2[j], &r1[j], 1, q);
#else
			r2[j] = DSA_ADD((r2[j]), (WALNUT_FIELD ^ r1[j]));
#endif
		}

#if WALNUT_SL == 256
		avxAdd128(&r2[j], &r1[j], 1, q128);
#endif
	}

	return 0;
}

static void process_digest(walnut_verify_ctx *c, const uint8_t *dgst)
{
	size_t i;

	AE_LOOKUP const uint8_t braiddata[] = {
		0x08, 0x07, 0x06, 0x05, 0x04, 0x03, 0x02, 0x01, 0x00,
			0x00, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18,
		0x08, 0x07, 0x06, 0x05, 0x04, 0x03, 0x02, 0x01,
			0x01, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18,
		0x08, 0x07, 0x06, 0x05, 0x04, 0x03, 0x02,
			0x02, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18,
		0x08, 0x07, 0x06, 0x05, 0x04, 0x03,
			0x03, 0x14, 0x15, 0x16, 0x17, 0x18,
		0x08, 0x07, 0x06, 0x05, 0x04,
			0x04, 0x15, 0x16, 0x17, 0x18,
		0x08, 0x07, 0x06, 0x05,
			0x05, 0x16, 0x17, 0x18,
		0x08, 0x07, 0x06,
			0x06, 0x17, 0x18,
		0x08, 0x07,
			0x07, 0x18,
		0x08,
			0x08,
	};

	// gendata value x -> (x+1)(x+3)(x+5)(x+7) encoder
	// example: x=0 -> 1357 encoder; x=1 -> 2468 encoder
	// for Bn, xmax=n-8
	 const uint8_t gendata[] = { 2, 1, 0, 1 };

	for(i=0; i<WALNUT_BRAID; i++) {
		c->encode[i][i] = 1;
	}

	for(i=0; i<WALNUT_BRAID; i++) {
		c->p[i] = i;
	}

	const uint8_t maxlength = (2 * WALNUT_BRAID) - 1;
	uint8_t reduced = 0;
	uint8_t word2 = gendata[0] + (2 * (dgst[0] >> 6));

	for(i=1; i<WALNUT_DGST_SIZ * 4; i++) {
		uint8_t word1;
		const uint8_t *b;
		uint8_t len;
		uint8_t shift;

		// last iteration's word2 is this iteration's word1
		word1 = word2;

		// point to the braid and index into it by last iteration's reduced
		// braiddata[word1 * (maxlength - word1)] = first generator of word1 braid
		b = braiddata + (word1 * (maxlength - word1)) + reduced;

		// wordlength = 2 * (WALNUT_BRAID - word - 1)
		len = (2 * (WALNUT_BRAID - word1 - 1)) - reduced;

		// calculate right shift for i-th 2-bit chunk
		shift = 2 * (3 - (i % 4));

		// 2-bit chunk (multiplied by 2 to align to proper braiddata row)
		word2 = (2 * ((dgst[i/4] >> shift) & 0x03));
		// gendata[i % sizeof(gendata)]: encoder (0 -> 1357, 1 -> 2468, etc.)
		word2 += gendata[i % sizeof(gendata)];

		// reduced = (min(word1length, word2length) / 2) - (1 if word1 >= word2; 0 else)
		word1 = WALNUT_BRAID - word1 - 1 - 1;
		reduced = WALNUT_BRAID - word2 - 1;
		if(word1 <= reduced) {
			reduced = word1;
		}

		// cut the reduced generators from the tail of the braid
		len -= reduced;
		walnut_emul(c, c->encode, b, len, 0);
	}

	// last 2-bit chunk
	const uint8_t *b;
	uint8_t len;

	b = braiddata + (word2 * (maxlength - word2)) + reduced;
	len = (2 * (WALNUT_BRAID - word2 - 1)) - reduced;

	walnut_emul(c, c->encode, b, len, 0);

}




extern int walnut_verify_init(const void *aekey, const void *dgst,
		walnut_verify_ctx *c)
{
	uint8_t i, j;

	aeLibReInit(WALNUT_FIELD);

	const GF_T *key = ((GF_T *)aekey)+1;
	const GF_T *smatrix = ((GF_T *)aekey)+1;

	//Skip past the T-values to point to the first element of the S matrix
	smatrix += WALNUT_BRAID;

	uint32_t cfg = HTON(((WALNUT_BRAID << 16) | (((WALNUT_FIELD >> 14) & 1) << 14) | (uint32_t)(ceil(log2(WALNUT_FIELD))))  << ((sizeof(GF_T) * 8) - (3 * 8)));

	enum {
		FIELD_MASK = ((uint32_t)(-1)) >> (32 - WALNUT_FIELD_BIT),
		BRAID_MASK = ((uint32_t)(-1)) >> (32 - WALNUT_BRAID_BIT)
	};

	if(memcmp(&cfg, aekey, sizeof(cfg))) {
		return -1;
	}

	memset(c, 0, sizeof(*c));
	//
	for(i=0; i<WALNUT_BRAID; i++) {
		//emul requires negative tvalues
		//in a prime field, -t = WALNUT_FIELD - t
		//because WALNUT_FIELD is all 1's from MSB to LSB, we can XOR instead
		ffelem tv = WALNUT_FIELD ^ NTOH(*key);
		key++;
#if WALNUT_SL == 256
		c->mp[i] = tv << 3;
		c->imp[i] = DSA_INVERSE(tv) << 3;
#else
		c->mp[i] = tv << 1;
		c->imp[i] = DSA_INVERSE(tv) << 1;
#endif
		//curr_bit += WALNUT_FIELD_BIT;
	}

	// skip the s matrix to get to the s' matrix
	key += (WALNUT_BRAID * (WALNUT_BRAID - 1)) + 1;

    for(i=0; i<WALNUT_BRAID; i++) {
		for(j=0; j<WALNUT_BRAID; j++) {

			if(i == WALNUT_BRAID-1) {
				if(j != WALNUT_BRAID-1)
					c->m.ansprime[j][i] = 0;
				else
				{
					c->m.ansprime[j][i] =NTOH(*key);
					key++;
				}
			} else {
				c->m.ansprime[j][i] = NTOH(*key);
				key++;
			}

		}
	}

#ifdef KAT_INTERMEDIATES
   //print_column_major_matrix(c->m.ansprime, "S prime matrix");
#endif

	memset(c->encode, 0, sizeof(c->encode));

	/* encode = E(dgst) */
	//I * encode(dgst)
	process_digest(c, dgst);

#ifdef KAT_INTERMEDIATES
    print_column_major_matrix(c->encode, "E(hash) * Identity Matrix");
#endif

	/* ans2 = (I * E(dgst)) \dot Pub_Sprime */
	mmul(c->d.ans2, c->encode, c->m.ansprime);

#ifdef KAT_INTERMEDIATES
    print_column_major_matrix(c->d.ans2, "E(hash) * Public Key");
#endif


	//F32 matrices end part-way through a byte (the remainder of which is purposefully zero'd out)
	//so we need to advance our pointer to the first bit of the permutation (round up to nearest 8)
	//curr_bit += (8 - (cu8rr_bit % 8)) & 0x07;

    AePackCtx pack;
	ae_pack_init(&pack, key, PACKED_PERMUTATION_LENGTH, 4);
	for(i=0; i<WALNUT_BRAID; i++) {
		c->p[i] = ae_unpack_entry(&pack);
		if (c->p[i] >= WALNUT_BRAID)
			return -2;
	}

	// revert curr_bit back to the s matrix location
	key = smatrix;

	for(i=0; i<WALNUT_BRAID; i++) {
		for(j=0; j<WALNUT_BRAID; j++) {

            if(i == WALNUT_BRAID-1) {
            	if(j != WALNUT_BRAID-1)
					c->m.ans1[j][i] = 0;
				else
				{
					c->m.ans1[j][i] =NTOH(*key);
					key++;
				}
            } else {
                c->m.ans1[j][i] = NTOH(*key);
				key++;
            }
        }
	}

#ifdef KAT_INTERMEDIATES
    //print_column_major_matrix(c->m.ans1, "S matrix");
#endif

	return 0;
}

extern int walnut_verify_update(walnut_verify_ctx *c, const void *sig,
		size_t len) 
{
	const uint8_t *s = sig;


	if(c->idx == 0 && len > 0) {
		c->count = *s << 8;

		/* Check that # of generators < 2^WALNUT_LENGTH_BIT */
		if(c->count & (uint16_t)(~0 << WALNUT_LENGTH_BIT)) {
		    return 0;
		}

		s++;
		len--;
		c->idx++;
	}

	if(c->idx == 1 && len > 0) {
		c->count |= *s;

		s++;
		len--;
		c->idx++;
	}

	if(c->idx >= 2 && len > 0) {
		enum {
			SIGBITS = WALNUT_BRAID_BIT + 1,
		};

		size_t generators = (len * 8 + c->numbits) / SIGBITS;
		uint8_t gen, i = 0, offset = 0;

		/* overflow */
		if(generators < len) {
			return 0;
		}

		if(generators > c->count) {
				generators = c->count;
		}

		if(generators) {

			c->count -= generators;

			// if there were unused generator bits from last signature chunk
			if(c->numbits) {
				offset = SIGBITS - c->numbits;

				// j is the leftover generator bits combined with
				// its corresponding bits from the front of s
				gen = (c->leftover << offset) | (*s >> (8 - offset));
				walnut_emul(c, c->m.ans1, &gen, ++i, 0);
			}

			// number of leftover bits
			c->numbits = (len * 8 + c->numbits) - (generators * SIGBITS);

			// leftover byte
			c->leftover = s[len - 1];

			walnut_emul(c, c->m.ans1, s, generators - i, 0x80 | offset);
		}


	}

#ifdef KAT_INTERMEDIATES
	print_column_major_matrix(c->m.ans1, "Public Key * Signature");
#endif

	return 0;
}

extern int walnut_verify_final(walnut_verify_ctx *c)
{
	int ret = memcmp(c->m.ans1[0], c->d.ans2[0], sizeof(c->m.ans1));

	return ret == 0 ? ret : -1;
}


