/*
 * primefields.c - implementation for computing in prime finite fields
 *
 * Copyright (C) 2015, SecureRF Corporation, All Rights Reserved.
 */

#include "primefields64.h"
#include <stdio.h>
#include <math.h>
#include "emmintrin.h"
#include "smmintrin.h"

static ffelem pf_add(const struct SrfFiniteField_* ff, ffelem a, ffelem b)
{

  unsigned long long x = (unsigned long long) a + b;

  if(x < ff->primepower)
	  return (ffelem) x;

  return (ffelem) (x - ff->primepower);

}

static inline ffelem mul_inv(ffelem a, const struct SrfFiniteField_* ff)
{
	ffelem b = ff->primepower;
	ffelem b0 =  ff->primepower;
	long t, q;
	long x0 = 0, x1 = 1;
	while (a > 1) {
		q = a / b;
		t = b, b = a % b, a = t;
		t = x0, x0 = x1 - q * x0, x1 = t;
	}
	return (ffelem) ((x1 + b0) % b0);
}

static ffelem pf_neg(const struct SrfFiniteField_* ff, ffelem a)
{
  return a ? ff->primepower - a : 0;
}

static ffelem pf_inverse(const struct SrfFiniteField_* ff, ffelem a)
{
  return a ? mul_inv(a, ff) : 0;
}
static ffelem pf_neginverse(const struct SrfFiniteField_* ff, ffelem a)
{
  return a ? ff->primepower - mul_inv(a, ff) : 0;
}

static ffelem pf_mult(const struct SrfFiniteField_* ff, ffelem a, ffelem b)
{
    ffelem hi, lo;

    // we want the "cutoff" between the lo/hi 61 bits to be at a 64 bit boundary (performance reasons)
    // shifting left three will do this, but note the lo 61 bit result will be offset left by 3
    a <<= 3;

#ifndef NO_MULTIPLY_64_BIT
#if __GNUC__ > 2 || \
              (__GNUC__ == 2 && (__GNUC_MINOR__ > 9 || \
                                 (__GNUC_MINOR__ == 9 && \
                                  __GNUC_PATCHLEVEL__ > 5)))
    __uint128_t test = (__uint128_t) a * (__uint128_t) b;
    __uint128_t *testptr = &test;

    hi = ((ffelem*)testptr)[1];
    lo = ((ffelem*)testptr)[0];

#else

    __asm__("mulq %3\n\t"
			: "=d" (hi),
			"=a" (lo)
			: "%a" (a),
			"rm" (b)
			: "cc" );

#endif

    return pf_add(ff, hi,lo >>= 3);
#else

    //We do this instead of hard coding it to 32 so that the bit shifts below don't cause errors if
	//compiled on a 32 bit platform (error will be because the bit shifts the width of the variable)
	const int halflength = (sizeof(ffelem) * 8) / 2;

	// if a = cx + j and b = dx + k, then
	// a * b = (cx + j)(dx + k) = (cd)x^2 + (ck + jd)x + (jk)
	// below: cd = x2, ck = x1a, jd = x1b, and jk = x0
	ffelem x2, x1a, x1b, x0;

	x2 = (a >> halflength) * (b >> halflength);
	x1a = (a >> halflength) * (b & 0xFFFFFFFF);
	x1b = (a & 0xFFFFFFFF) * (b >> halflength);
	x0 = (a & 0xFFFFFFFF) * (b & 0xFFFFFFFF);

	lo = x0 + (x1a << halflength);
	hi = (lo < (x1a << halflength));

	lo += (x1b << halflength);
	hi += (lo < (x1b << halflength));

	hi += x2 + (x1a >> halflength) + (x1b >> halflength);

	// fix the left 3 offset by shifting right 3
	lo >>= 3;

	return pf_add(ff, hi, lo);
#endif

}

static ffelem pf_mult_no_shift(const struct SrfFiniteField_* ff, ffelem a, ffelem b)
{
    ffelem hi, lo;

    // we want the "cutoff" between the lo/hi 61 bits to be at a 64 bit boundary (performance reasons)
    // shifting left three will do this, but note the lo 61 bit result will be offset left by 3


#ifndef NO_MULTIPLY_64_BIT
#if __GNUC__ > 2 || \
              (__GNUC__ == 2 && (__GNUC_MINOR__ > 9 || \
                                 (__GNUC_MINOR__ == 9 && \
                                  __GNUC_PATCHLEVEL__ > 5)))
    __uint128_t test = (__uint128_t) a * (__uint128_t) b;
    __uint128_t *testptr = &test;

    hi = ((ffelem*)testptr)[1];
    lo = ((ffelem*)testptr)[0];

#else

    __asm__("mulq %3\n\t"
			: "=d" (hi),
			"=a" (lo)
			: "%a" (a),
			"rm" (b)
			: "cc" );

#endif

    return pf_add(ff, hi,lo >>= 3);
#else

    //We do this instead of hard coding it to 32 so that the bit shifts below don't cause errors if
	//compiled on a 32 bit platform (error will be because the bit shifts the width of the variable)
	const int halflength = (sizeof(ffelem) * 8) / 2;

	// if a = cx + j and b = dx + k, then
	// a * b = (cx + j)(dx + k) = (cd)x^2 + (ck + jd)x + (jk)
	// below: cd = x2, ck = x1a, jd = x1b, and jk = x0
	ffelem x2, x1a, x1b, x0;

	x2 = (a >> halflength) * (b >> halflength);
	x1a = (a >> halflength) * (b & 0xFFFFFFFF);
	x1b = (a & 0xFFFFFFFF) * (b >> halflength);
	x0 = (a & 0xFFFFFFFF) * (b & 0xFFFFFFFF);

	lo = x0 + (x1a << halflength);
	hi = (lo < (x1a << halflength));

	lo += (x1b << halflength);
	hi += (lo < (x1b << halflength));

	hi += x2 + (x1a >> halflength) + (x1b >> halflength);

	// fix the left 3 offset by shifting right 3
	lo >>= 3;

	return pf_add(ff, hi, lo);
#endif

}
static ffelem pf_reduce(const struct SrfFiniteField_* ff, ffelem a)
{
  return a % ff->primepower;
}

__m128i Multiply64Bit(__m128i a, __m128i b)
{
	__m128i ax0_ax1_ay0_ay1 = a;
	__m128i bx0_bx1_by0_by1 = b;

    // i means ignored

    __m128i ax1_i_ay1_i = _mm_shuffle_epi32(ax0_ax1_ay0_ay1, _MM_SHUFFLE(3, 3, 1, 1));
    __m128i bx1_i_by1_i = _mm_shuffle_epi32(bx0_bx1_by0_by1, _MM_SHUFFLE(3, 3, 1, 1));

    __m128i ax0bx0_ay0by0 = _mm_mul_epi32(ax0_ax1_ay0_ay1, bx0_bx1_by0_by1);
    __m128i ax0bx1_ay0by1 = _mm_mul_epi32(ax0_ax1_ay0_ay1, bx1_i_by1_i);
    __m128i ax1bx0_ay1by0 = _mm_mul_epi32(ax1_i_ay1_i, bx0_bx1_by0_by1);

    __m128i ax0bx1_ay0by1_32 = _mm_slli_epi64(ax0bx1_ay0by1, 32);
    __m128i ax1bx0_ay1by0_32 = _mm_slli_epi64(ax1bx0_ay1by0, 32);

    return _mm_add_epi64(ax0bx0_ay0by0, _mm_add_epi64(ax0bx1_ay0by1_32, ax1bx0_ay1by0_32));
}

#define DEFINE_FF(num) \
  { num, pf_neg, pf_inverse, pf_neginverse, pf_add, pf_mult, pf_reduce, pf_mult_no_shift}

static SrfFiniteField ffFields[] = {
#ifdef ENABLE_FIELD_M61
  DEFINE_FF(2305843009213693951),
#endif
};
#define NUM_FIELDS (sizeof(ffFields)/sizeof(*ffFields))


extern const SrfFiniteField* ffGetPrimeField64(ffelem primepower)
{
  int i;
  for (i = 0; i < NUM_FIELDS; i++) {
    if (ffFields[i].primepower == primepower)
      return &ffFields[i];
  }
  return NULL;
}

// #endif
